import os
from typing import Dict

from magma.backend.coreir.insert_coreir_wires import insert_coreir_wires
from magma.backend.mlir.compile_to_mlir import compile_to_mlir
from magma.backend.mlir.compile_to_mlir_opts import CompileToMlirOpts
from magma.backend.mlir.mlir_to_verilog import (
    MlirToVerilogOpts,
    mlir_to_verilog,
)
from magma.circuit import DefineCircuitKind
from magma.common import slice_opts
from magma.compiler import Compiler
from magma.passes.clock import wire_clocks
from magma.passes.elaborate_tuples import elaborate_tuples
from magma.passes.when import run_when_passes
from magma.passes.raise_logs_as_exceptions import raise_logs_as_exceptions_pass


class MlirCompiler(Compiler):
    def __init__(self, main: DefineCircuitKind, basename: str, opts: Dict):
        # TODO(rsetaluri): Make this a better error.
        assert "basename" not in opts
        opts["basename"] = basename
        self._compile_to_mlir_opts = slice_opts(
            opts, CompileToMlirOpts, keep=True
        )
        self._mlir_to_verilog_opts = slice_opts(
            opts, MlirToVerilogOpts, keep=True
        )
        super().__init__(main, basename, opts)

    def suffix(self):
        if self.opts.get("output_verilog", False):
            if self.opts.get("sv", False):
                return "sv"
            return "v"
        return "mlir"

    def run_pre_uniquification_passes(self):
        if self._compile_to_mlir_opts.flatten_all_tuples:
            elaborate_tuples(self.main)
        # NOTE(leonardt): when passes must happen after any
        # passes that modify the circuit.  This is because passes
        # could introduce more conditional logic, or they could
        # trigger elaboration on values used in existing coditiona
        # logic (which modifies the when builder).
        run_when_passes(
            self.main,
            self._compile_to_mlir_opts.flatten_all_tuples,
            self._compile_to_mlir_opts.emit_when_assertions,
        )

    def _run_passes(self):
        raise_logs_as_exceptions_pass(self.main)
        insert_coreir_wires(self.main, flatten=False)
        wire_clocks(self.main)

    def compile(self):
        self._run_passes()
        with open(f"{self.basename}.mlir", "w") as f:
            compile_to_mlir(
                self.main, sout=f, opts=self._compile_to_mlir_opts)
        if not self.opts.get("output_verilog", False):
            return
        outfile = (
            os.devnull
            if self.opts.get("split_verilog", False)
            else f"{self.basename}.{self.suffix()}"
        )
        with open(f"{self.basename}.mlir", "r") as fi:
            with open(outfile, "w") as fo:
                mlir_to_verilog(fi, fo, opts=self._mlir_to_verilog_opts)



from magma.bits import Bits
from magma.bitutils import clog2
from magma.clock import Enable
from magma.clock_io import ClockIO
from magma.generator import Generator2
from magma.interface import IO
from magma.t import In, Out, Kind


class MultiportMemory(Generator2):
    def __init__(
            self,
            height: int,
            T: Kind,
            num_read_ports: int = 1,
            num_write_ports: int = 1,
            has_read_enable: bool = False
    ):
        if num_read_ports < 1:
            raise ValueError("At least one read port is required")
        if num_write_ports < 0:
            raise ValueError("Number of write ports must be non-negative")

        self.num_read_ports = num_read_ports
        self.has_read_enable = has_read_enable
        self.num_write_ports = num_write_ports
        self.T = T
        self.height = height

        addr_width = clog2(height)

        self.io = ClockIO()
        for i in range(num_read_ports):
            self.io += IO(**{
                f"RADDR_{i}": In(Bits[addr_width]),
                f"RDATA_{i}": Out(T)
            })
            if has_read_enable:
                self.io += IO(**{f"RE_{i}": In(Enable)})
        for i in range(num_write_ports):
            self.io += IO(**{
                f"WADDR_{i}": In(Bits[addr_width]),
                f"WDATA_{i}": In(T),
                f"WE_{i}": In(Enable)
            })
        self.primitive = True



import magma as m
from magma.testing import check_files_equal


class _MyMux(m.Generator2):
    cls_level_var = "I like being a mux"

    def __init__(self, width, height):
        self.width = width
        self.height = height
        sel_bits = m.bitutils.clog2(height)
        self.name = f"MyMux{width}x{height}"
        args = {f"I{i}": m.In(m.Bits[width]) for i in range(height)}
        args["O"] = m.Out(m.Bits[width])
        args["S"] = m.In(m.Bits[sel_bits])
        self.io = m.IO(**args)

    def to_string(self):
        return f"{_MyMux.cls_level_var} of {self.width}x{self.height}"


def test_type_relations():
    assert issubclass(_MyMux, m.Generator2)
    assert issubclass(_MyMux, m.DefineCircuitKind)
    MyMux4x4 = _MyMux(4, 4)  # circuit defn.
    assert isinstance(MyMux4x4, m.DefineCircuitKind)
    assert isinstance(MyMux4x4, _MyMux)
    assert issubclass(MyMux4x4, m.Circuit)

    insts = []

    class _(m.Circuit):
        mux = MyMux4x4()
        insts.append(mux)

    mux = insts[0]
    assert isinstance(mux, MyMux4x4)
    assert isinstance(mux, m.Circuit)


def test_properties():
    MyMux4x4 = _MyMux(4, 4)
    assert MyMux4x4.name == "MyMux4x4"
    assert MyMux4x4.width == 4
    assert MyMux4x4.height == 4
    assert MyMux4x4.to_string() == "I like being a mux of 4x4"


def test_compilation():

    class _Top(m.Circuit):
        io = m.IO(x=m.In(m.Bit), y=m.In(m.Bit), z=m.Out(m.Bit))
        mux = _MyMux(1, 2)()
        mux.I0[0] @= io.x
        mux.I1[0] @= io.y
        mux.S @= 0
        io.z @= mux.O[0]

    m.compile("build/TopGen", _Top, output="coreir")
    assert check_files_equal(__file__, "build/TopGen.json", "gold/TopGen.json")


def test_cache():
    MyMux4x8 = _MyMux(4, 8)
    MyMux4x8_other = _MyMux(4, 8)
    assert MyMux4x8 is MyMux4x8_other
    MyMux4x16 = _MyMux(4, 16)
    assert MyMux4x16 is not MyMux4x8


def test_cache_miss():

    # Different class, with same parameters; internals don't matter.
    class _MyOtherMux(m.Generator2):
        def __init__(self, width, height):
            pass

    MyMux4x8 = _MyMux(4, 8)
    MyOtherMux4x8 = _MyOtherMux(4, 8)
    assert MyMux4x8 is not MyOtherMux4x8


def test_no_cache():

    class _MyGen(m.Generator2):
        _cache_ = False

        def __init__(self):
            self.io = m.IO(x=m.In(m.Bit))

    my_gen = _MyGen()
    my_gen_other = _MyGen()
    assert my_gen is not my_gen_other
    assert repr(my_gen) == repr(my_gen_other)


def test_subclass_generator_instance():
    Base = m.Register(m.Bit)

    class MyRegister(Base):
        io = Base.io
        verilog = "Foo"

    assert isinstance(MyRegister, m.Register)



from magma.when import (
    when,
    elsewhen,
    otherwise,
    WhenBlock,
    ElseWhenBlock,
    OtherwiseBlock
)


def _check_to_split(value, outputs, to_split):
    source = value.trace()
    for elem in source.root_iter():
        if any(elem is output for output in outputs):
            to_split.append(source)
            return
    if value.has_children() and value.has_elaborated_children():
        for _, child in value.enumerate_children():
            _check_to_split(child, outputs, to_split)


def _get_builder_ports(builder, names):
    """Filter out removed ports."""
    return [getattr(builder, name) for name in names if hasattr(builder, name)]


def _find_values_to_split(builder):
    """Detect output values that feed into inputs."""
    to_split = []
    outputs = _get_builder_ports(builder, builder.output_to_name.values())
    inputs = _get_builder_ports(builder, builder.input_to_name.values())
    for value in inputs:
        _check_to_split(value, outputs, to_split)
    return to_split


def _emit_new_when_assign(value, driver_map, curr_block):
    """Reconstruct when logic in new set of blocks."""
    if isinstance(curr_block, WhenBlock):
        new_block = when(curr_block._info.condition)
    elif isinstance(curr_block, ElseWhenBlock):
        new_block = elsewhen(curr_block._info.condition)
    elif isinstance(curr_block, OtherwiseBlock):
        new_block = otherwise()
    with new_block:
        if curr_block in driver_map:
            for driver in driver_map[curr_block]:
                value @= driver
        for child in curr_block.children():
            _emit_new_when_assign(value, driver_map, child)
    for _elsewhen in curr_block.elsewhen_blocks():
        _emit_new_when_assign(value, driver_map, _elsewhen)
    if curr_block.otherwise_block:
        _emit_new_when_assign(value, driver_map, curr_block.otherwise_block)


def _build_driver_map(drivee):
    # driver_map stores the driver for each context that drivee is driven.
    driver_map = {}
    for ctx in drivee._wired_when_contexts:
        wires = ctx.get_conditional_wires_for_drivee(drivee)
        driver_map[ctx] = (wire.driver for wire in wires)
    return driver_map


def split_when_cycles(builder, defn):
    to_split = _find_values_to_split(builder)
    for value in to_split:
        driving = value.driving()
        driver_map = _build_driver_map(driving[0])
        for drivee in driving:
            drivee.unwire()
            root = next(iter(driver_map.keys())).root
            _emit_new_when_assign(drivee, driver_map, root)



import abc
from collections import defaultdict
import contextlib
import dataclasses
import functools
import itertools
import operator
from typing import Any, Iterable, Optional, Set, Tuple, Union

from magma.debug import get_debug_info, debug_info as DebugInfo
from magma.ref import get_ref_inst


@functools.lru_cache(None)
def _bit_type():
    # NOTE(leonardt): Circular dependency.
    from magma.bit import Bit
    return Bit


@functools.lru_cache(None)
def _abstract_register_type():
    # NOTE(rsetaluri): Circular dependency.
    from magma.primitives.register import AbstractRegister
    return AbstractRegister


@functools.lru_cache(None)
def _enable_type():
    # NOTE(rsetaluri): Circular dependency.
    from magma.clock import Enable
    return Enable


def _contains_tuple(T):
    # NOTE(leonardt): Circular dependency.
    from magma.type_utils import contains_tuple
    return contains_tuple(T)


def _inline_verilog2(*args, **kwargs):
    # NOTE(leonardt): Circular dependency.
    from magma.inline_verilog2 import inline_verilog2
    inline_verilog2(*args, **kwargs)


class WhenSyntaxError(SyntaxError):
    pass


class NoPrecedingWhenError(WhenSyntaxError):
    def __init__(self, type):
        super().__init__(f"Cannot use {type} without a previous when")


class ElsewhenWithoutPrecedingWhenError(NoPrecedingWhenError):
    def __init__(self):
        super().__init__("elsewhen")


class OtherwiseWithoutPrecedingWhenError(NoPrecedingWhenError):
    def __init__(self):
        super().__init__("otherwise")


@dataclasses.dataclass(frozen=True)
class ConditionalWire:
    drivee: Any
    driver: Any


class _BlockBase(contextlib.AbstractContextManager):
    def __init__(self, parent: Optional['WhenBlock']):
        self._parent = parent
        self._children = list()
        self._conditional_wires = list()
        self._default_drivers = dict()

    def spawn(self, info: 'WhenBlockInfo') -> 'WhenBlock':
        child = WhenBlock(self, info)
        self._children.append(child)
        return child

    def get_conditional_wires_for_drivee(self, i):
        return list(
            filter(lambda x: x.drivee is i, self._conditional_wires)
        )

    def remove_conditional_wire(self, i):
        self._conditional_wires = list(
            filter(lambda x: x.drivee is not i, self._conditional_wires)
        )
        if self.root._default_drivers.pop(i, None) is not None:
            self.root.builder.remove_default_driver(i)

    def add_conditional_wire(self, i, o):
        self._conditional_wires.append(ConditionalWire(i, o))
        builder = self.root.builder
        if i.driven() and i not in builder.output_to_index:
            driver = i.value()
            # Keep wired when contexts because in this case, the driver could
            # have been wired in a different when context (so the default value
            # comes from the result of a previous when context)
            i.unwire(driver, keep_wired_when_contexts=True)
            self.root.add_default_driver(i, driver)
        self.root.builder.add_drivee(i)
        self.root.builder.add_driver(o)

    @property
    @abc.abstractmethod
    def root(self) -> 'WhenBlock':
        raise NotImplementedError()

    @abc.abstractmethod
    def new_elsewhen_block(self, info: 'ElseWhenBlockInfo'):
        raise NotImplementedError()

    @abc.abstractmethod
    def new_otherwise_block(self):
        raise NotImplementedError()

    @property
    @abc.abstractmethod
    def condition(self) -> Optional:
        raise NotImplementedError()

    @abc.abstractmethod
    def elsewhen_blocks(self) -> Iterable['_BlockBase']:
        raise NotImplementedError()

    @property
    @abc.abstractmethod
    def otherwise_block(self) -> Optional['_BlockBase']:
        raise NotImplementedError()

    def conditional_wires(self) -> Iterable[ConditionalWire]:
        yield from self._conditional_wires

    def default_drivers(self) -> Iterable[ConditionalWire]:
        if self.root is not self:
            raise RuntimeError("Can only get default drivers from root")
        return (
            ConditionalWire(k, v)
            for k, v in self._default_drivers.items()
        )

    def get_default_drivers_dict(self) -> Iterable[ConditionalWire]:
        if self.root is not self:
            raise RuntimeError("Can only get default drivers from root")
        return self._default_drivers

    def get_default_driver(self, drivee) -> Any:
        if self.root is not self:
            raise RuntimeError("Can only get default driver from root")
        return self._default_drivers[drivee]

    def add_default_driver(self, drivee, driver):
        if self.root is not self:
            raise RuntimeError("Can only add default driver to root")
        self._default_drivers[drivee] = driver
        self.root.builder.add_default_driver(drivee, driver)

    def children(self) -> Iterable['_BlockBase']:
        yield from self._children

    def __enter__(self):
        if self.condition is not None:
            self.root.builder.add_condition(self.condition)
        _set_curr_block(self)
        return self

    def _add_reg_enables(self):
        """
        If we wire the input of a register, check if there is an enable port
        and wire to 1 if found.

        This will overwrite an existing wiring if the user already has done it.

        The warning raised will indicate to the user that the explicit wiring
        is unnecessary and promote the idiom of not explicitly wiring enables.
        """
        conditional_wires = list(self._conditional_wires)
        # We need to keep track of Enable values we've wired and make sure to
        # only wire them once. Otherwise, an elaborated register input value may
        # be wired multiple times.
        # NOTE(leonardt): If we wanted to allow users to wire Enable values
        # explicitly, we could lookup whether a wire to enable exists in
        # conditional_wires (which would also avoid multiple wirings).
        seen = set()
        for wire in conditional_wires:
            inst = get_ref_inst(wire.drivee.name)
            if (
                    inst is None
                    or not isinstance(type(inst), _abstract_register_type())
            ):
                continue
            ce_port = type(inst).get_enable(inst)
            if (
                    ce_port is None
                    # Explicitly wired in this block, skip.
                    or self.get_conditional_wires_for_drivee(ce_port)
                    or (
                        ce_port.driven()
                        and not ce_port.driven_implicitly_by_when
                    )
                    or ce_port in seen
            ):
                continue
            ce_port.wire(1, driven_implicitly_by_when=True)
            seen.add(ce_port)

    @abc.abstractmethod
    def _get_exit_block(self):
        """Returns the block that should be set as the current block when
        __exit__ is called.
        """
        raise NotImplementedError()

    def __exit__(self, exc_type, exc_value, traceback):
        self._add_reg_enables()
        _set_curr_block(self._get_exit_block())
        _set_prev_block(self)


BlockBase = _BlockBase


def get_all_blocks(
        block: _BlockBase,
        include_self: bool = True,
) -> Iterable[_BlockBase]:
    if include_self:
        yield block
    sub_blocks = itertools.chain(
        block.children(),
        block.elsewhen_blocks(),
        () if block.otherwise_block is None else (block.otherwise_block,),
    )
    for sub_block in sub_blocks:
        yield from get_all_blocks(sub_block)


@dataclasses.dataclass
class WhenBlockInfo:
    # NOTE(leonardt): We can't use Bit here because of circular dependency, so
    # instead we enforce it in the public constructors (e.g. m.when).
    condition: Any


@dataclasses.dataclass
class ElseWhenBlockInfo(WhenBlockInfo):
    pass


class WhenBlock(_BlockBase):
    def __init__(
            self,
            parent: Optional[_BlockBase],
            info: WhenBlockInfo,
            debug_info: Optional[DebugInfo] = None,
    ):
        super().__init__(parent)
        self._info = info
        self._elsewhens = list()
        self._otherwise = None
        self._builder = None
        if self._parent is None:
            # NOTE(rsetaluri): Circular dependency.
            from magma.primitives.when import WhenBuilder
            self._builder = WhenBuilder(self)
            if debug_info is not None:
                self._builder.debug_info = debug_info

    def new_elsewhen_block(self, info: 'ElseWhenBlockInfo'):
        block = ElseWhenBlock(self, info)
        self._elsewhens.append(block)
        return block

    def new_otherwise_block(self):
        if self._otherwise is not None:
            raise WhenSyntaxError()
        block = OtherwiseBlock(self)
        self._otherwise = block
        return block

    @property
    def builder(self) -> Any:
        if self._builder is None:
            raise AttributeError()
        return self._builder

    @property
    def root(self) -> 'WhenBlock':
        if self._parent is None:
            return self
        return self._parent.root

    @property
    def condition(self) -> Any:
        return self._info.condition

    def elsewhen_blocks(self) -> Iterable['_BlockBase']:
        yield from self._elsewhens

    @property
    def otherwise_block(self) -> Optional['_BlockBase']:
        return self._otherwise

    def __enter__(self):
        this = _BlockBase.__enter__(self)
        assert this is self
        _set_prev_block(None)
        return self

    def _get_exit_block(self):
        return self._parent


class ElseWhenBlock(_BlockBase):
    def __init__(self, parent: Optional[_BlockBase], info: ElseWhenBlockInfo):
        super().__init__(parent)
        self._info = info

    def new_elsewhen_block(self, info: 'ElseWhenBlockInfo'):
        return self._parent.new_elsewhen_block(info)

    def new_otherwise_block(self):
        return self._parent.new_otherwise_block()

    @property
    def root(self) -> WhenBlock:
        return self._parent.root

    @property
    def condition(self) -> Any:
        return self._info.condition

    def elsewhen_blocks(self) -> Iterable['_BlockBase']:
        yield from ()

    @property
    def otherwise_block(self) -> None:
        return None

    def _get_exit_block(self):
        return self._parent._get_exit_block()


class OtherwiseBlock(_BlockBase):
    def new_elsewhen_block(self, _):
        raise ElsewhenWithoutPrecedingWhenError()

    def new_otherwise_block(self):
        raise OtherwiseWithoutPrecedingWhenError()

    @property
    def root(self) -> WhenBlock:
        return self._parent.root

    @property
    def condition(self) -> None:
        return None

    def elsewhen_blocks(self) -> Iterable['_BlockBase']:
        yield from ()

    @property
    def otherwise_block(self) -> None:
        return None

    def _get_exit_block(self):
        return self._parent._get_exit_block()


_curr_block = None
_prev_block = None


@contextlib.contextmanager
def no_when():
    block = _get_curr_block()
    _set_curr_block(None)
    yield
    _set_curr_block(block)


@contextlib.contextmanager
def temp_when(temp_ctx):
    block = _get_curr_block()
    _set_curr_block(temp_ctx)
    yield
    _set_curr_block(block)


def _get_curr_block() -> Optional[_BlockBase]:
    global _curr_block
    return _curr_block


def _set_curr_block(curr_block: Optional[_BlockBase]):
    global _curr_block
    _curr_block = curr_block


def _reset_curr_block():
    _set_curr_block(None)


def _get_prev_block() -> Optional[_BlockBase]:
    global _prev_block
    return _prev_block


def _set_prev_block(prev_block: Optional[_BlockBase]):
    global _prev_block
    _prev_block = prev_block


def _reset_prev_block():
    _set_prev_block(None)


get_curr_block = _get_curr_block


def _reset_context():
    _reset_curr_block()
    _reset_prev_block()
    global _ASSERT_COUNTER
    _ASSERT_COUNTER = itertools.count()


reset_context = _reset_context


def _get_else_block(block: _BlockBase) -> Optional[_BlockBase]:
    if isinstance(block, OtherwiseBlock):
        return None
    if isinstance(block, ElseWhenBlock):
        # TODO(rsetaluri): We could augment ElseWhenBlock to keep track of its
        # index in the parent block, or alternatively keep an explicit pointer
        # to the next else/otherwise block in each ElseWhenBlock.
        parent_elsewhen_blocks = list(block._parent.elsewhen_blocks())
        index = parent_elsewhen_blocks.index(block) + 1
        if index == len(parent_elsewhen_blocks):
            return block._parent.otherwise_block
        return parent_elsewhen_blocks[index]
    if isinstance(block, WhenBlock):
        try:
            else_block = next(block.elsewhen_blocks())
        except StopIteration:
            pass
        else:
            return else_block
        return block.otherwise_block


_Op = Union[ConditionalWire, _BlockBase]


def _get_then_ops(block: _BlockBase) -> Iterable[_Op]:
    yield from block.conditional_wires()
    yield from block.children()


def _get_else_ops(block: _BlockBase) -> Iterable[_Op]:
    if isinstance(block, OtherwiseBlock):
        return _get_then_ops(block)
    if isinstance(block, ElseWhenBlock):
        return (block,)
    raise TypeError(block)


def _get_assignees_and_latches(ops: Iterable[_Op]) -> Tuple[Set, Set]:
    """Compute the assigned values and latches implied by a block.

    Note: this is a conservative check that requires values are assigned in
    every case.  For example, there may be an unreachable case where a value is
    not assigned, but that requires symbolic evaluation to discover.  In
    general, symbolic evaluation is difficult, but may be viable in the
    hardware domain where we may have less issues with path explosion,
    aliasing, arrays, etc...

    Parameters
    ----------
    ops : Iterable[_Op]
        A sequence of ops (assignment or if-then-else block)

    Returns
    -------
    Set[Type]
        The values assigned in this block by any operation.

    Set[Type]
        The latches implied by this block, i.e. values which are not assigned in
        all branches.
    """
    assignees, latches = set(), set()
    for op in ops:
        if isinstance(op, ConditionalWire):
            assignees.add(op.drivee)
            continue
        if isinstance(op, _BlockBase):
            then_ops = _get_then_ops(op)
            then_assignees, then_latches = _get_assignees_and_latches(
                then_ops
            )
            else_block = _get_else_block(op)
            if else_block is None:
                latches.update(then_latches.difference(assignees))
                latches.update(then_assignees.difference(assignees))
                continue
            else_ops = _get_else_ops(else_block)
            else_assignees, else_latches = _get_assignees_and_latches(
                else_ops
            )
            latches.update(then_latches.difference(assignees))
            latches.update(else_latches.difference(assignees))
            latches.update(
                (
                    then_assignees
                    .symmetric_difference(else_assignees)
                    .difference(assignees)
                )
            )
            assignees.update(then_assignees.union(else_assignees))
            # Filter out any latches that were created in this when context.
            latches = set(filter(
                lambda x: x.enclosing_when_context is not op,
                latches
            ))
            continue
        raise TypeError(op)
    return assignees, latches


def find_inferred_latches(block: _BlockBase) -> Set:
    if not (isinstance(block, WhenBlock) and block.root is block):
        raise TypeError("Can only find inferred latches on root when block")
    ops = tuple(block.default_drivers()) + (block,)
    _, latches = _get_assignees_and_latches(ops)
    return latches


# We use a global counter so assertion names are unique across the design
# (not just per module).  This is because the downstream verilog tools
# (e.g. Jasper) treat each assertion as a unique entity (rather than scoped in
# the hierarchy), so they are expected to be unique like module names.
_ASSERT_COUNTER = itertools.count()
_ASSERT_TEMPLATE = (
    "WHEN_ASSERT_{id}: assert property (({cond}) |-> ({drivee} == {driver}));"
)


def _get_builder_port(value, builder):
    """
    Get reference to corresponding builder output port.  With the bulk
    reconstruction logic, this may refer to a child of a builder output.
    """
    port = builder.check_existing_derived_ref(
        value, builder.output_to_name, builder.output_to_index
    )
    if port is None:
        port = getattr(builder, builder.output_to_name[value])
    return port


def _emit_when_assert(
    cond, drivee, driver, builder, assignment_map, flatten_all_tuples
):
    if _contains_tuple(type(drivee)) and flatten_all_tuples:
        # Since tuples are elaborated in verilog, we emit an assert for the
        # leaf values.
        for x, y in zip(drivee, driver):
            _emit_when_assert(
                cond, x, y, builder, assignment_map, flatten_all_tuples
            )
        return

    port = _get_builder_port(drivee, builder)
    if not port.wired():
        return  # port was discarded, do not need to emit assert

    # Track the conditions a value is assigned in for the default driver logic.
    assignment_map[port].append(cond)
    port = builder.get_when_assert_wire(port)
    _inline_verilog2(
        _ASSERT_TEMPLATE,
        cond=cond,
        drivee=port,
        driver=driver,
        id=next(_ASSERT_COUNTER)
    )


def _make_cond(block, precond):
    """Prepends the precond to the block condition if it exists."""
    if precond is not None:
        if block.condition is not None:
            return precond & block.condition
        return precond
    assert block.condition is not None
    return block.condition


def _make_else_cond(block, precond):
    """Invert the current condition, if there's a precond, append it."""
    if precond is not None:
        return precond & ~block.condition
    return ~block.condition


def _emit_default_driver_asserts(
    block, builder, assignment_map, flatten_all_tuples
):
    for wire in list(block.root.default_drivers()):
        port = _get_builder_port(wire.drivee, builder)
        assigned_conds = assignment_map[port]
        if not assigned_conds:
            continue
        cond = ~functools.reduce(operator.or_, assigned_conds)
        _emit_when_assert(
            cond,
            wire.drivee,
            wire.driver,
            builder,
            defaultdict(list),
            flatten_all_tuples
        )


def emit_when_assertions(
    block,
    builder,
    flatten_all_tuples,
    precond=None,
    assignment_map=defaultdict(list)
):
    """
    For each drivee in a conditional wire, track the conditions where it is
    assigned using assignment_map.  This is used for the default driver logic
    (emit an assert for the case when none of the relevant conditions are
    true).
    """
    cond = _make_cond(block, precond)
    for conditional_wire in block.conditional_wires():
        _emit_when_assert(
            cond,
            conditional_wire.drivee,
            conditional_wire.driver,
            builder,
            assignment_map,
            flatten_all_tuples
        )

    for child in block.children():
        # Children inherit this block's cond as a precond.
        emit_when_assertions(
            child, builder, flatten_all_tuples, cond, assignment_map
        )
    else_block = _get_else_block(block)
    if else_block:
        else_cond = _make_else_cond(block, precond)
        emit_when_assertions(
            else_block, builder, flatten_all_tuples, else_cond, assignment_map
        )
    if block is block.root:
        _emit_default_driver_asserts(
            block, builder, assignment_map, flatten_all_tuples
        )


def when(cond):
    if not isinstance(cond, _bit_type()):
        raise TypeError(f"Invalid when cond {cond}, should be Bit")
    info = WhenBlockInfo(cond)
    curr_block = _get_curr_block()
    if curr_block is None:
        return WhenBlock(None, info, get_debug_info(3))
    return curr_block.spawn(info)


def elsewhen(cond):
    if not isinstance(cond, _bit_type()):
        raise TypeError(f"Invalid elsewhen cond {cond}, should be Bit")
    info = ElseWhenBlockInfo(cond)
    prev_block = _get_prev_block()
    if prev_block is None:
        raise ElsewhenWithoutPrecedingWhenError()
    return prev_block.new_elsewhen_block(info)


def otherwise():
    prev_block = _get_prev_block()
    if prev_block is None:
        raise OtherwiseWithoutPrecedingWhenError()
    return prev_block.new_otherwise_block()



import magma as m
import hwtypes as ht
from magma.type_utils import to_hwtypes


def test_bit_to_ht():
    assert to_hwtypes(m.Bit) is ht.Bit


def test_bits_to_ht():
    assert to_hwtypes(m.Bits[5]) is ht.BitVector[5]


def test_sint_to_ht():
    assert to_hwtypes(m.SInt[5]) is ht.SIntVector[5]


def test_uint_to_ht():
    assert to_hwtypes(m.UInt[5]) is ht.UIntVector[5]


def test_tuple_to_ht():
    assert (to_hwtypes(m.Tuple[m.Bit, m.Bits[5]]) is
            ht.Tuple[ht.Bit, ht.BitVector[5]])


def test_product_to_ht():
    class T(m.Product):
        x = m.Bit
        y = m.Bits[5]

    # Need them to be the same name so we just rename the reference
    MT = T

    class T(ht.Product):
        x = ht.Bit
        y = ht.BitVector[5]

    HTT = T

    assert to_hwtypes(MT) == HTT



import pytest

import magma as m
from magma.testing.utils import has_warning


class _MixedTuple(m.Product):
    x = m.In(m.Bit)
    y = m.Out(m.Bit)


def _make_io():
    return m.IO(
        I=m.In(m.Bit),
        I0=m.In(m.Array[8, m.Bits[10]]),
        mixed=_MixedTuple,
        O=m.Out(m.Bit),
        O3=m.Out(m.Array[4, m.Bits[6]]),
    )


def _get_inputs(obj):
    yield obj.O
    yield obj.mixed.y
    yield from obj.O3


def _get_outputs(obj):
    yield obj.I
    yield from obj.I0
    yield obj.mixed.x


def _check_open_not_implemented(cls):
    with pytest.raises(NotImplementedError):
        cls.open()


def test_stubify_ckt():

    class _Foo(m.Circuit):
        io = _make_io()

    m.stubify(_Foo, stubbifier=m.zero_stubbifier)

    assert m.isdefinition(_Foo)
    _check_open_not_implemented(_Foo)
    drivers = (port.trace() for port in _get_inputs(_Foo))
    assert all(driver.const() and int(driver) == 0 for driver in drivers)


def test_decorator():

    @m.circuit_stub(stubbifier=m.zero_stubbifier)
    class _Foo(m.Circuit):
        io = _make_io()

    assert m.isdefinition(_Foo)
    _check_open_not_implemented(_Foo)
    drivers = (port.trace() for port in _get_inputs(_Foo))
    assert all(driver.const() and int(driver) == 0 for driver in drivers)


def test_subclass():

    class _Foo(m.CircuitStub):
        io = _make_io()

    assert m.isdefinition(_Foo)
    _check_open_not_implemented(_Foo)
    drivers = (port.trace() for port in _get_inputs(_Foo))
    assert all(driver.const() and int(driver) == 0 for driver in drivers)


def test_io():

    class _Foo(m.Circuit):
        io = _make_io()
        m.stubify(io, stubbifier=m.zero_stubbifier)

    assert m.isdefinition(_Foo)
    drivers = (port.trace() for port in _get_inputs(_Foo))
    assert all(driver.const() and int(driver) == 0 for driver in drivers)


def test_instance():

    class _Foo(m.Circuit):
        io = _make_io()
        # NOTE(rsetaluri): This stubbification is not the one we care about. It
        # is just here to facilitate stubbing the instance of this circuit
        # below.
        m.stubify(io)

    class _(m.Circuit):
        io = m.IO()
        inst = _Foo()
        m.stubify(inst.interface, stubbifier=m.zero_stubbifier)
        # NOTE(rsetaluri): We have to use map() here because of quirks with
        # doing list comprehension inside of class bodies.
        drivers = map(lambda p: p.trace(), _get_outputs(inst))
        assert all(map(lambda d: d.const() and int(d) == 0, drivers))


@pytest.mark.parametrize(
    "stubbifier",
    (
        m.zero_stubbifier,
        m.no_override_driven_zero_stubbifier
    )
)
def test_partially_driven(caplog, stubbifier):

    class _Foo(m.Circuit):
        io = m.IO(O=m.Out(m.Bits[2]))
        io.O @= 1
        m.stubify(io, stubbifier=stubbifier)

    expected_warning = (stubbifier is m.zero_stubbifier)

    if expected_warning:
        assert has_warning(caplog)


def test_clocks():

    class _Foo(m.Circuit):
        io = m.IO()
        reg = m.Register(m.Bit)()
        m.stubify(reg.interface)

    assert _Foo.reg.I.driven()
    driver = _Foo.reg.I.trace()
    assert driver.const() and int(driver) == 0
    assert not _Foo.reg.CLK.driven()


def test_nested_clocks():

    class Clk(m.Product):
        clk = m.Clock

    class _Bar(m.Circuit):
        io = m.IO(I=m.In(m.Bit), clk=m.In(Clk))

    class _Foo(m.Circuit):
        io = m.IO(I=m.In(m.Bit), clk=m.In(Clk))
        bar = _Bar()
        m.stubify(bar.interface)

    assert _Foo.bar.I.driven()
    driver = _Foo.bar.I.trace()
    assert driver.const() and int(driver) == 0
    assert not _Foo.bar.clk.driven()



"""
Based on
https://github.com/chipsalliance/chisel3/blob/master/src/main/scala/chisel3/util/Decoupled.scala
(missing support for `flow` and `pipe` parameters).
"""
from magma.clock_io import ClockIO
from magma.conversions import enable
from magma.generator import Generator2
from magma.interface import IO
from magma.mantle.counter import Counter
from magma.primitives.memory import Memory
from magma.primitives.register import Register
from magma.types.handshake import DeqIO, EnqIO
from magma.t import Kind


class Queue(Generator2):
    """
    Queue implementation using a magma Memory primitive as a register file.
    Write (enqueue) and read (dequeue) counters point to the approriate
    addresses in memory.

    Since enq_ptr == deq_ptr could mean the queue is full or empty, this
    implementation uses an extra bit to track when the queue is full.
    """

    def __init__(self, entries: int, T: Kind):
        assert entries >= 0
        self.io = IO(
            # NOTE(leonardt): Note that the direction of enq/deq are flipped
            # since this is a client (consumer)
            enq=DeqIO[T],
            deq=EnqIO[T]
        ) + ClockIO()

        # Data storage
        ram = Memory(entries, T)()

        # Read/write pointers
        enq_ptr = Counter(entries, has_enable=True)()
        deq_ptr = Counter(entries, has_enable=True)()

        # Since the pointers can match when it's empty or full, we use an extra
        # bit to track when it may be full (there's been a write without a
        # read)
        maybe_full = Register(init=False, has_enable=True)()

        ptr_match = enq_ptr.O == deq_ptr.O
        # Empty/full determined by pointers matching and maybe_full bit
        empty = ptr_match & ~maybe_full.O
        full = ptr_match & maybe_full.O

        # deq data is valid when not empty
        self.io.deq.valid @= ~empty
        # enq is ready when not full
        self.io.enq.ready @= ~full

        # do enq/deq when ready/valid are high
        do_enq = self.io.enq.fired()
        do_deq = self.io.deq.fired()

        # write enq data when do_enq
        ram.write(self.io.enq.data, enq_ptr.O, enable(do_enq))

        # Increment pointers on read/write
        enq_ptr.CE @= enable(do_enq)
        deq_ptr.CE @= enable(do_deq)

        # Set maybe full when enq without deq
        maybe_full.I @= enable(do_enq)
        maybe_full.CE @= enable(do_enq != do_deq)

        # Deq data from storage
        self.io.deq.data @= ram[deq_ptr.O]



from magma.bits import Bits
from magma.conversions import bits, zext_to
from magma.generator import Generator2
from magma.interface import IO
from magma.t import In, Out


class Decoder(Generator2):
    """Decodes n-bit value into a one hot encoding (2^n bits).
    """

    def __init__(self, n: int):
        self.io = IO(I=In(Bits[n]), O=Out(Bits[2**n]))
        self.io.O @= bits(1, 2**n) << zext_to(self.io.I, 2**n)


def decode(x: Bits) -> Bits:
    n = len(x)
    return Bits[2**n](1) << zext_to(x, 2**n)



import itertools
import pathlib
import pytest
import string

import magma as m
from magma.bind2 import (
    is_bound_module,
    is_bound_instance,
    maybe_get_bound_instance_info,
)
from magma.passes.passes import CircuitPass
import magma.testing


def _assert_compilation(ckt, basename, suffix, opts):
    # Specialize gold file for cwd.
    cwd = pathlib.Path(__file__).parent
    with open(f"{cwd}/gold/{basename}.{suffix}.tpl", "r") as f_tpl:
        with open(f"{cwd}/gold/{basename}.{suffix}", "w") as f_out:
            tpl = string.Template(f_tpl.read())
            f_out.write(tpl.substitute(cwd=cwd))
    m.compile(f"build/{basename}", ckt, **opts)
    assert m.testing.check_files_equal(
        __file__,
        f"build/{basename}.{suffix}",
        f"gold/{basename}.{suffix}")


@pytest.mark.parametrize(
    "backend,split_verilog",
    itertools.product(("mlir",), (False, True))
)
def test_basic(backend, split_verilog):

    class Top(m.Circuit):
        io = m.IO(I=m.In(m.Bit), O=m.Out(m.Bit))
        io.O @= ~io.I

    class TopBasicAsserts(m.Circuit):
        name = f"TopBasicAsserts_{backend}"
        io = m.IO(I=m.In(m.Bit), O=m.In(m.Bit), other=m.In(m.Bit))
        io.I.unused()
        io.O.unused()
        io.other.unused()

    m.bind2(Top, TopBasicAsserts, Top.I)

    assert not is_bound_module(Top)
    assert is_bound_module(TopBasicAsserts)

    basename = f"test_bind2_basic"
    suffix = "mlir" if backend == "mlir" else "v"
    opts = {
        "output": backend,
        "use_native_bind_processor": True,
    }
    if split_verilog:
        opts.update({"split_verilog": True})
        basename = basename + "_split_verilog"
    _assert_compilation(Top, basename, suffix, opts)


@pytest.mark.parametrize(
    "backend,flatten_all_tuples,inst_attr",
    (
        ("mlir", False, False),
        ("mlir", True, False),
        ("mlir", True, True),
    )
)
def test_xmr(backend, flatten_all_tuples, inst_attr):

    class T(m.Product):
        x = m.Bit
        y = m.Bit

    U = m.Tuple[m.Bit, m.Bits[8]]

    class Bottom(m.Circuit):
        io = m.IO(I=m.In(T), O=m.Out(T), x=m.Out(U))
        io.O @= io.I
        io.x[0] @= 0
        io.x[1] @= 0

    class Middle(m.Circuit):
        io = m.IO(I=m.In(T), O=m.Out(T))
        io.O @= Bottom(name="bottom")(io.I)[0]

    class Top(m.Circuit):
        io = m.IO(I=m.In(T), O=m.Out(T))
        if inst_attr:
            middle = Middle(name="middle")
            O = middle(io.I)
        else:
            O = Middle(name="middle")(io.I)
        io.O @= O

    class TopXMRAsserts(m.Circuit):
        name = f"TopXMRAsserts_{backend}"
        io = m.IO(I=m.In(T), O=m.In(T), a=m.In(T), b=m.In(m.Bit), c=m.In(m.Bit))
        io.I.unused()
        io.O.unused()
        io.a.unused()
        io.b.unused()
        io.c.unused()

    m.bind2(
        Top,
        TopXMRAsserts,
        Top.middle.bottom.O,
        Top.middle.bottom.I.x,
        Top.middle.bottom.x[0],
    )

    basename = "test_bind2_xmr"
    if flatten_all_tuples:
        basename += "_flatten_all_tuples"
    suffix = "mlir" if backend == "mlir" else "v"
    opts = {
        "output": backend,
        "use_native_bind_processor": True,
        "flatten_all_tuples": flatten_all_tuples,
    }
    _assert_compilation(Top, basename, suffix, opts)


def test_generator():

    class Logic(m.Generator2):
        def __init__(self, width=None):
            T = m.Bit if width is None else m.Bits[width]
            self.io = io = m.IO(I=m.In(T), O=m.Out(T))
            io.O @= ~io.I

    class LogicAsserts(m.Generator2):
        def __init__(self, dut, width=None):
            T = m.Bit if width is None else m.Bits[width]
            self.width = width
            self.io = io = m.IO(I=m.In(T), O=m.In(T), other=m.In(m.Bit))
            m.inline_verilog("{I} {O} {other}", I=io.I, O=io.O, other=io.other)
            self.bind2_args = [m.bits(dut.I)[0]]

    m.bind2(Logic, LogicAsserts)

    class Top(m.Circuit):
        T = m.Bits[2]
        io = m.IO(I=m.In(T), O=m.Out(T))
        I = m.bits(list(map(lambda x: Logic()()(x), io.I)))
        io.O @= Logic(2)()(I)

    opts = {
        "output": "mlir",
        "use_native_bind_processor": True,
    }
    _assert_compilation(Top, "test_bind2_generator", "mlir", opts)

    class _CheckLogicAssertsAreBoundModulesPass(CircuitPass):
        def __call__(self, defn):
            if not isinstance(defn, LogicAsserts):
                return
            assert is_bound_module(defn)

    assert not is_bound_module(Top)
    _CheckLogicAssertsAreBoundModulesPass(Top).run()


@pytest.mark.parametrize(
    "backend,split_verilog",
    itertools.product(("mlir",), (False, True))
)
def test_compile_guard(backend, split_verilog):

    class Top(m.Circuit):
        io = m.IO(I=m.In(m.Bit), O=m.Out(m.Bit))
        io.O @= ~io.I

    class TopCompileGuardAsserts(m.Circuit):
        name = f"TopCompileGuardAsserts_{backend}"
        io = m.IO(I=m.In(m.Bit), O=m.In(m.Bit), other=m.In(m.Bit))
        io.I.unused()
        io.O.unused()
        io.other.unused()

    with m.compile_guard("ASSERT_ON"):
        m.bind2(Top, TopCompileGuardAsserts, Top.I)

    assert not is_bound_module(Top)
    assert is_bound_module(TopCompileGuardAsserts)
    inst = Top.instances[-1]
    assert is_bound_instance(inst)
    compile_guard_infos = maybe_get_bound_instance_info(inst).compile_guards
    assert len(compile_guard_infos) == 1
    assert compile_guard_infos[0] is not None
    assert compile_guard_infos[0].condition_str == "ASSERT_ON"

    basename = f"test_bind2_compile_guard"
    suffix = "mlir" if backend == "mlir" else "v"
    opts = {
        "output": backend,
        "use_native_bind_processor": True,
    }
    if split_verilog:
        opts.update({"split_verilog": True})
        basename = basename + "_split_verilog"
    _assert_compilation(Top, basename, suffix, opts)



from magma.backend.mlir.hw import hw
from magma.backend.mlir.mlir import push_block
from magma.backend.mlir.sv import sv


def make_mem_reg(ctx, name, N, T):
    mem_type = hw.InOutType(hw.ArrayType((N,), T))
    mem = ctx.new_value(mem_type)
    sv.RegOp(name=name, results=[mem])
    return mem


def make_mem_read(ctx, target, value, has_enable, name="read_reg"):
    if not has_enable:
        sv.ReadInOutOp(operands=[target], results=[value])
        return None, None
    reg_out = ctx.new_value(target.type.T)
    sv.ReadInOutOp(operands=[target], results=[reg_out])
    reg = ctx.new_value(target.type)
    sv.RegOp(name=name, results=[reg])
    sv.ReadInOutOp(operands=[reg], results=[value])
    return reg, reg_out


def emit_conditional_assign(target, value, en):
    with push_block(sv.IfOp(operands=[en]).then_block):
        sv.PAssignOp(operands=[target, value])


def make_index_op(ctx, value, idx):
    result = ctx.new_value(hw.InOutType(value.type.T.T))
    sv.ArrayIndexInOutOp(operands=[value, idx], results=[result])
    return result


def collect_multiport_memory_operands(
    operands, start_idx, num_ports, num_operands_per_port
):
    """Collect flat list of operands for read or write ports.
    start_idx: offset in `operands` list to start from
    num_ports: number of ports to collect
    num_operands_per_port: number of operands per port
                           (e.g. 3 for waddr, wdata, wen)
    """
    port_operands = []
    curr_idx = start_idx
    for i in range(num_ports):
        port_operands.append(
            operands[curr_idx:curr_idx + num_operands_per_port]
        )
        curr_idx += num_operands_per_port
    return port_operands


def make_multiport_memory_index_ops(ctx, ports, mem):
    """For each port, emit an array index op and return a list of results."""
    return [make_index_op(ctx, mem, port[0]) for port in ports]


def make_multiport_memory_read_ops(
        ctx, read_results, read_ports, read_ports_out
):
    """If ren, emit an intermediate register to hold read value."""
    read_targets = []
    for i, (target, port) in enumerate(zip(read_results, read_ports)):
        has_en = len(port) == 2
        read_reg, read_temp = make_mem_read(
            ctx, target, read_ports_out[i], has_en, f"read_reg_{i}"
        )
        if has_en:
            read_targets.append((read_reg, read_temp, port[1]))
    return read_targets


def emit_conditional_assigns(targets):
    for target, data, en in targets:
        emit_conditional_assign(target, data, en)



import pytest
import magma as m
import hwtypes as ht


def _assert_identical(x, y):
    assert all(xi is yi for xi, yi in zip(x, y))


def test_concat_basic():
    x = m.Bit(1)
    y = m.Bits[1](1)
    o0 = m.concat(x, y)
    o1 = m.concat(m.bits(x), y)
    _assert_identical(o0, o1)
    _assert_identical(o0, [x, y[0]])


def test_concat_bit():
    x = ht.Bit(True)
    y = True
    z = m.Bits[1](1)
    o = m.concat(x, y, z)
    assert o.const()
    _assert_identical(o, [m.Bit.VCC] * 3)


def test_ext():
    x0 = m.SInt[1](1)

    x1 = m.zext_by(x0, 2)
    x2 = m.zext_to(x0, 3)
    assert isinstance(x1, m.SInt)
    assert isinstance(x2, m.SInt)
    assert x1.const()
    _assert_identical(x1, x2)
    _assert_identical(x1, [m.Bit.VCC, m.Bit.GND, m.Bit.GND])

    x3 = m.sext_by(x0, 2)
    x4 = m.sext_to(x0, 3)
    assert isinstance(x3, m.SInt)
    assert isinstance(x4, m.SInt)
    assert x3.const()
    _assert_identical(x3, x4)
    _assert_identical(x3, [m.Bit.VCC] * 3)


def test_concat_type_error():
    with pytest.raises(TypeError):
        m.concat(object(), object())


@pytest.mark.parametrize('op', [m.uint, m.sint])
def test_convert_extend(op):
    x = m.Bits[5]()
    o = op(x, 32)
    assert len(o) == 32
    if op is m.sint:  # check sext logic
        hi = o[5:]
        _assert_identical(hi, [x[4]] * len(hi))


@pytest.mark.parametrize(
    'Targs,Tout',
    (
        ((m.Bits[1], m.SInt[1]), m.Bits),
        ((m.SInt[1], m.SInt[1]), m.SInt),
        ((m.UInt[1], m.UInt[1]), m.UInt),
        ((m.SInt[1], m.UInt[1]), m.Bits),
        ((m.Bit, m.Bits[1]), m.Bits),
    )
)
def test_concat_output_type(Targs, Tout):
    args = (Targ() for Targ in Targs)
    o = m.concat(*args)
    assert isinstance(o, Tout)


@pytest.mark.parametrize('T', [m.UInt[4], m.SInt[4]])
def test_bits_upcast(T):
    assert type(m.bits(T())) is m.Bits[4]


def test_bit_upcast():
    x = m.Enable()
    assert type(m.bit(x)) is m.Bit


@pytest.mark.parametrize(
    'T,conversion,expected_T',
    [
        (m.Bit, m.clock, m.Clock),
        (m.Bits[4], m.uint, m.UInt),
        (m.UInt[4], m.bits, m.Bits),
        (m.Reset, m.bit, m.Bit)
    ]
)
def test_name(T, conversion, expected_T):
    x = T(name="x")
    y = conversion(x, name="y")
    assert str(y) == 'y'
    assert isinstance(y, expected_T)



from magma.passes.tsort import tsort


def test_basic():
    graph = [(2, []),
             (5, [11]),
             (11, [2, 9, 10]),
             (7, [11, 8]),
             (9, []),
             (10, []),
             (8, [9]),
             (3, [10, 8])]
    sorted_ = tsort(graph)

    # Check topo. sorted property of sorted_.
    indices = {k: i for i, (k, _) in enumerate(sorted_)}
    for k, neighbors in sorted_:
        index = indices[k]
        for neighbor in neighbors:
            assert indices[neighbor] < index

    # Explicitly check expectation of sorted list.
    assert sorted_ == [(2, []),
                       (9, []),
                       (10, []),
                       (8, [9]),
                       (3, [10, 8]),
                       (11, [2, 9, 10]),
                       (7, [11, 8]),
                       (5, [11]),]



import dataclasses
import io
from typing import Optional

import circt
import circt.passmanager

from magma.backend.mlir.errors import MlirCompilerError


@dataclasses.dataclass
class MlirToVerilogOpts:
    split_verilog: bool = False
    split_verilog_directory: Optional[str] = None


def mlir_to_verilog(
        istream: io.TextIOBase,
        ostream: io.TextIOBase,
        opts: MlirToVerilogOpts = MlirToVerilogOpts(),
):
    ir = istream.read()
    with circt.ir.Context() as ctx:
        circt.register_dialects(ctx)
        module = circt.ir.Module.parse(ir)
        passes = [
            "lower-seq-to-sv",
            "canonicalize",
            "hw.module(hw-cleanup)",
            "hw.module(prettify-verilog)",
        ]
        pass_string = ",".join(passes)
        pm = circt.passmanager.PassManager.parse(
            f"builtin.module({pass_string})"
        )
        pm.run(module.operation)
    if opts.split_verilog:
        directory = opts.split_verilog_directory or "."
        circt.export_split_verilog(module, directory)
    else:
        circt.export_verilog(module, ostream)



import pytest

import magma as m
from magma.value_utils import (
    ValueVisitor,
    ArraySelector,
    TupleSelector,
    make_selector,
    fill,
)


class _Prod(m.Product):
    x = m.Array[5, m.Bits[10]]
    y = m.Array[10, m.Bit]


def test_value_visitor():
    class _Visitor(ValueVisitor):
        def __init__(self):
            self.values = []

        def visit_Array(self, value):
            self.values.append(("Array", value))
            ValueVisitor.generic_visit(self, value)

        def visit_Tuple(self, value):
            self.values.append(("Tuple", value))
            ValueVisitor.generic_visit(self, value)

        def visit_Digital(self, value):
            self.values.append(("Digital", value))
            ValueVisitor.generic_visit(self, value)

        def visit_Bits(self, value):
            self.values.append(("Bits", value))
            ValueVisitor.generic_visit(self, value)

    T = m.Array[10, _Prod]
    v = _Visitor()
    value = T()
    v.visit(value)
    expected = [("Array", value)]
    for t in value:
        expected.append(("Tuple", t))
        expected.append(("Array", t.x))
        for tt in t.x:
            expected.append(("Bits", tt))
            for ttt in tt:
                expected.append(("Digital", ttt))
        expected.append(("Array", t.y))
        for tt in t.y:
            expected.append(("Digital", tt))

    assert len(expected) == len(v.values)
    assert all(e_str == v_str and e_val is v_val
               for ((e_str, e_val), (v_str, v_val)) in zip(expected, v.values))


def test_selector():
    T = m.Array[10, _Prod]

    class _Foo(m.Circuit):
        name = "Foo"
        io = m.IO(I=m.In(T))

    selector = ArraySelector(
        index=0,
        child=TupleSelector(
            key="x",
            child=ArraySelector(
                index=0,
                child=None)))

    assert str(selector) == "[0].x[0]"
    assert selector.select(_Foo.I) is _Foo.I[0].x[0]


@pytest.mark.parametrize("fill_value", (True, False))
def test_fill(fill_value):
    S = m.AnonProduct[{"x": m.Bits[8], "y": m.Bit}]
    T = m.AnonProduct[{"s": S, "u": m.Array[4, m.Bits[6]]}]

    t = T()
    fill(t, fill_value)

    value = t.value()
    assert value is not None

    assert value.s.const()
    assert value.s.x.const()
    assert value.s.y.const()
    assert int(value.s.x) == (0 if not fill_value else 255)
    assert int(value.s.y) == (0 if not fill_value else 1)

    assert value.u.const()
    for u_i in value.u:
        assert u_i.const()
        assert int(u_i) == (0 if not fill_value else 63)


def test_selector_stop_at():
    class T(m.Product):
        x = m.Bit
        y = m.Tuple[m.Bit, m.Bits[8]]

    class _Foo(m.Circuit):
        name = "Foo"
        io = m.IO(I=m.In(T))

        selector = make_selector(io.I.y[1][0], stop_at=io.I.y)
        assert selector.select(io.I.y) is io.I.y[1][0]



import pytest
import magma as m
from magma.testing import check_files_equal


def test_verilog_field_uniquify():
    # https://github.com/phanrahan/magma/issues/330.
    class HalfAdder(m.Circuit):
        name = "HalfAdder"
        io = m.IO(A=m.In(m.Bit), B=m.In(m.Bit), S=m.Out(m.Bit), C=m.Out(m.Bit))
        io.verilog = """\
            assign S = A ^ B;
            assign C = A & B;\
        """


def test_uniquify_equal():
    class foo(m.Circuit):
        name = "foo"
        io = m.IO(I=m.In(m.Bit), O=m.Out(m.Bit))
        m.wire(io.I, io.O)

    class bar(m.Circuit):
        name = "foo"
        io = m.IO(I=m.In(m.Bit), O=m.Out(m.Bit))
        m.wire(io.I, io.O)

    class top(m.Circuit):
        name = "top"
        io = m.IO(I=m.In(m.Bit), O=m.Out(m.Bit))
        curr = io.I
        for circ in (foo, bar):
            inst = circ()
            m.wire(inst.I, curr)
            curr = inst.O
        m.wire(curr, io.O)

    assert hash(repr(foo)) == hash(repr(bar))

    m.compile("build/uniquify_equal", top, output="coreir")
    assert check_files_equal(__file__,
                             "build/uniquify_equal.json",
                             "gold/uniquify_equal.json")


def test_uniquify_unequal():
    class foo(m.Circuit):
        name = "foo"
        io = m.IO(I=m.In(m.Bit), O=m.Out(m.Bit))
        m.wire(io.I, io.O)

    class bar(m.Circuit):
        name = "foo"
        io = m.IO(I=m.In(m.Bits[2]), O=m.Out(m.Bits[2]))
        m.wire(io.I, io.O)

    class top(m.Circuit):
        name = "top"
        io = m.IO(I=m.In(m.Bit), O=m.Out(m.Bit))
        foo_inst = foo()
        m.wire(io.I, foo_inst.I)
        bar_inst = bar()
        m.wire(foo_inst.O, bar_inst.I[0])
        m.wire(foo_inst.O, bar_inst.I[1])
        m.wire(bar_inst.O[0], io.O)

    assert hash(repr(foo)) != hash(repr(bar))

    m.compile("build/uniquify_unequal", top, output="coreir")
    assert check_files_equal(__file__,
                             "build/uniquify_unequal.json",
                             "gold/uniquify_unequal.json")


def test_key_error():
    default_port_mapping = {
        "I": "in",
        "I0": "in0",
        "I1": "in1",
        "O": "out",
        "S": "sel",
    }

    class I(m.Product):
        data = m.Array[2, m.Bits[6]]
        sel = m.Bits[m.bitutils.clog2(2)]

    class _Mux(m.Circuit):
        name = f"coreir_commonlib_mux{2}x{6}"
        io = m.IO(I=m.In(I), O=m.Out(m.Bits[6]))
        coreir_name = "muxn"
        coreir_lib = "commonlib"
        coreir_genargs = {"width": 6, "N": 2}
        renamed_ports = default_port_mapping

    class Mux2x6(m.Circuit):
        name = "Mux2x6"
        io = m.IO(I0=m.In(m.Bits[6]), I1=m.In(
        m.Bits[6]), S=m.In(m.Bit), O=m.Out(m.Bits[6]))

        mux = _Mux()
        m.wire(io.I0, mux.I.data[0])
        m.wire(io.I1, mux.I.data[1])
        m.wire(io.S, mux.I.sel[0])
        m.wire(mux.O, io.O)

    class MuxWrapper_2_6(m.Circuit):
        name = "MuxWrapper_2_6"
        io = m.IO(I=m.Array[2, m.In(
        m.Bits[6])], S=m.In(m.Bits[1]), O=m.Out(m.Bits[6]))
        Mux2x6_inst0 = Mux2x6()
        m.wire(io.I[0], Mux2x6_inst0.I0)
        m.wire(io.I[1], Mux2x6_inst0.I1)
        m.wire(io.S[0], Mux2x6_inst0.S)
        m.wire(Mux2x6_inst0.O, io.O)

    class MuxWrapper_2_6_copy(m.Circuit):
        name = "MuxWrapper_2_6"
        io = m.IO(I=m.Array[2, m.In(
        m.Bits[6])], S=m.In(m.Bits[1]), O=m.Out(m.Bits[6]))
        Mux2x6_inst0 = Mux2x6()
        m.wire(io.I[0], Mux2x6_inst0.I0)
        m.wire(io.I[1], Mux2x6_inst0.I1)
        m.wire(io.S[0], Mux2x6_inst0.S)
        m.wire(Mux2x6_inst0.O, io.O)

    class MuxWithDefaultWrapper_2_6_19_0(m.Circuit):
        name = "MuxWithDefaultWrapper_2_6_19_0"
        io = m.IO(I=m.Array[2, m.In(
        m.Bits[6])], S=m.In(m.Bits[19]), O=m.Out(m.Bits[6]))
        MuxWrapper_2_6_inst0 = MuxWrapper_2_6()
        MuxWrapper_2_6_inst1 = MuxWrapper_2_6_copy()
        m.wire(io.I, MuxWrapper_2_6_inst0.I)
        m.wire(io.I, MuxWrapper_2_6_inst1.I)
        m.wire(io.S[0], MuxWrapper_2_6_inst0.S[0])
        m.wire(io.S[0], MuxWrapper_2_6_inst1.S[0])
        m.wire(MuxWrapper_2_6_inst1.O, io.O)

    top = MuxWithDefaultWrapper_2_6_19_0
    m.compile(f"build/{top.name}", top, output="coreir")
    assert check_files_equal(__file__,
                             f"build/{top.name}.json",
                             "gold/uniquification_key_error_mux.json")


def test_uniquify_verilog():
    [foo0] = m.define_from_verilog("""
module foo(input I, output O);
    assign O = I;
endmodule""")
    [foo1] = m.define_from_verilog("""
module foo(input II, output OO);
    assign OO = II;
endmodule""")
    assert repr(foo0) != repr(foo1)
    class top(m.Circuit):
        name = "top"
        io = m.IO(I=m.In(m.Bit), O=m.Out(m.Bit))
        foo0_inst = foo0()
        foo1_inst = foo1()
        m.wire(io.I, foo0_inst.I)
        m.wire(foo0_inst.O, foo1_inst.II)
        m.wire(foo1_inst.OO, io.O)

    with pytest.raises(Exception) as pytest_e:
        m.compile(f"top", top, output="coreir")
        assert False
    assert pytest_e.type is Exception
    assert pytest_e.value.args == \
        ("Can not rename a verilog wrapped file",)


def test_hash_verilog():
    [foo0] = m.define_from_verilog("""
module foo(input I, output O);
    assign O = I;
endmodule""")
    class foo1(m.Circuit):
        name = "foo"
        io = m.IO(I=m.In(m.Bit), O=m.Out(m.Bit))
        is_definition = True

    class top(m.Circuit):
        name = "top"
        io = m.IO(I=m.In(m.Bit), O=m.Out(m.Bit))
        foo0_inst = foo0()
        foo1_inst = foo1()
        m.wire(io.I, foo0_inst.I)
        m.wire(foo0_inst.O, foo1_inst.I)
        m.wire(foo1_inst.O, io.O)

    assert repr(foo0) == repr(foo1)

    # Run the uniquification pass as a mechanism to check that foo0 and foo1
    # hash to two different things even though they have the same repr.
    pass_ = m.UniquificationPass(top, None)
    pass_.run()
    foo_seen = pass_.seen["foo"]
    assert len(foo_seen) == 2
    for v in foo_seen.values():
        assert len(v) == 1
    expected_ids_ = {id(v[0]) for v in foo_seen.values()}
    ids_ = {id(foo0), id(foo1)}
    assert expected_ids_ == ids_


def test_same_verilog():
    """
    This test checks that if we have multiple verilog-wrapped circuits with the
    same name and same source, the uniquification pass does *not* try to perform
    a rename. As it does not need to.
    """
    def _generate_foo():
        return m.define_from_verilog("""
module foo(input i, output o);
    assign o = i;
endmodule""")[0]

    class _Cell0(m.Circuit):
        io = m.IO(I=m.In(m.Bit), O=m.Out(m.Bit))

        foo = _generate_foo()()
        foo.i <= io.I
        io.O <= foo.o

    class _Cell1(m.Circuit):
        io = m.IO(I=m.In(m.Bit), O=m.Out(m.Bit))

        foo = _generate_foo()()
        foo.i <= io.I
        io.O <= foo.o

    class _Top(m.Circuit):
        io = m.IO(I=m.In(m.Bits[2]), O=m.Out(m.Bits[2]))

        cell0 = _Cell0()
        cell1 = _Cell1()
        cell0.I <= io.I[0]
        cell1.I <= io.I[1]
        io.O[0] <= cell0.O
        io.O[1] <= cell1.O

    # Check that uniq. pass runs successfully.
    pass_ = m.UniquificationPass(_Top, None)
    pass_.run()


def test_multiple_renamed():
    def _gen_foo(width):
        class Foo(m.Circuit):
            io = m.IO(I=m.In(m.Bits[width]), O=m.Out(m.Bits[width]))

            io. O <= io.I

        return Foo

    Foo0 = _gen_foo(2)
    Foo1 = _gen_foo(3)
    Foo2 = _gen_foo(3)

    class _Top(m.Circuit):
        io = m.IO(
            I0=m.In(m.Bits[2]),
            I1=m.In(m.Bits[3]),
            I2=m.In(m.Bits[3]),
            O0=m.Out(m.Bits[2]),
            O1=m.Out(m.Bits[3]),
            O2=m.Out(m.Bits[3]),
        )

        foo0 = Foo0()
        foo1 = Foo1()
        foo2 = Foo2()

        foo0.I <= io.I0
        io.O0 <= foo0.O

        foo1.I <= io.I1
        io.O1 <= foo1.O

        foo2.I <= io.I2
        io.O2 <= foo2.O

    BASENAME = "uniquify_multiple_rename"
    m.compile(f"build/{BASENAME}", _Top, output="coreir")
    assert check_files_equal(__file__,
                             f"build/{BASENAME}.json",
                             f"gold/{BASENAME}.json")


def test_reset_names():

    def _make_foo(width: int):

        class _Foo(m.Circuit):
            name = "Foo"
            io = m.IO(I=m.In(m.Bits[width]), O=m.Out(m.Bits[width]))
            io.O @= io.I

        return _Foo

    Foo1 = _make_foo(1)
    Foo2 = _make_foo(2)

    class _Top1(m.Circuit):
        name = "Top1"
        io = m.IO(I=m.In(m.Bits[3]), O=m.Out(m.Bits[3]))
        io.O[0:1] @= Foo1()(io.I[0:1])
        io.O[1:3] @= Foo2()(io.I[1:3])

    # Because _Top1 has instances of 2 different circuits with the name "Foo",
    # we expect one of them to be uniquified to "Foo_unq1".
    basename = "uniquify_reset_names_1"
    m.compile(f"build/{basename}", _Top1, output="coreir")
    assert check_files_equal(
        __file__, f"build/{basename}.json", f"gold/{basename}.json")

    # Names of circuits should be reset after compilation, even though they were
    # uniquified temporarily.
    assert Foo1.name == "Foo"
    assert Foo2.name == "Foo"

    # We make sure to clear the CoreIR context.
    m.frontend.coreir_.ResetCoreIR()
    m.generator.reset_generator_cache()

    Foo3 = _make_foo(3)

    # Define a new top circuit which will contain instances of *3* different
    # circuits each with the name "Foo". If names were not reset after
    # uniquification above, then there will be a conflict. In particular, Foo3
    # will be renamed to "Foo_unq1" even though Foo2 is already named "Foo_unq1"
    # (*if* names were *not* reset).
    class _Top2(m.Circuit):
        name = "Top2"
        io = m.IO(I=m.In(m.Bits[6]), O=m.Out(m.Bits[6]))
        io.O[0:1] @= Foo1()(io.I[0:1])
        io.O[1:3] @= Foo2()(io.I[1:3])
        io.O[3:6] @= Foo3()(io.I[3:6])

    basename = "uniquify_reset_names_2"
    m.compile(f"build/{basename}", _Top2, output="coreir")
    assert check_files_equal(
        __file__, f"build/{basename}.json", f"gold/{basename}.json")



import magma as m

from magma.logging import (
    root_logger, stage_logger, unstage_logger, staged_logs
)
from magma.testing.utils import has_info


class _NamedObject:
    def __init__(self, name: str):
        self.name = name

    def __str__(self):
        return self.name


def test_basic_staging(caplog):
    msg = _NamedObject("foo")
    logger = root_logger()

    # First do a sanity check that staged logging without any modifications to
    # the staged object logs as expected.
    stage_logger()
    logger.info(msg)
    unstage_logger()
    assert caplog.records[-1].message == "foo"

    # Next check that modifying the object *after* staging & logging, but
    # *before* unstaging, logs the modified obejct.
    stage_logger()
    logger.info(msg)
    msg.name = "bar"
    unstage_logger()
    assert caplog.records[-1].message == "bar"


def test_nested_staging(caplog):
    msg = _NamedObject("foo")
    logger = root_logger()

    # Check that interleaving logging within unstaging works as expected. After
    # the first logger is unstaged, the log should be flushed to console, but
    # the outer logger should remain staged. Therefore, we should see one log
    # with "foo" right after the first unstage, and one log with "bar" after the
    # next unstage.
    stage_logger()
    stage_logger()
    logger.info(msg)
    unstage_logger()
    assert caplog.records[-1].message == "foo"
    logger.info(msg)
    msg.name = "bar"
    unstage_logger()
    assert caplog.records[-1].message == "bar"


def test_staged_logs_context_manager(caplog):
    msg = _NamedObject("foo")
    logger = root_logger()

    # Check that the staged_logs() context manager appropriately stages logs,
    # *and* the target returned from staged_logs() is the list of staged logs
    # itself.
    with staged_logs() as logs:
        logger.info(msg)
        msg.name = "bar"
    assert len(logs) == 1
    assert logs[0][2] is msg
    assert caplog.records[-1].message == "bar"



import magma as m
import operator
from common import DeclareAnd
import pytest
from magma.testing import check_files_equal


@pytest.mark.parametrize("width", [None, 3])
@pytest.mark.parametrize("output", ["verilog", "coreir"])
@pytest.mark.parametrize("op", [operator.imatmul, operator.le])
def test_assign(width, output, op):
    T = m.util.BitOrBits(width)

    class Circ(m.Circuit):
        name = f"test_assign_operator_{width}_{output}"
        io = m.IO(a=m.In(T), b=m.In(T), c=m.Out(T))
        and2 = DeclareAnd(width)()
        op(and2.I0, io.a)
        op(and2.I1, io.b)
        op(io.c, and2.O)

    m.compile(f"build/{Circ.name}", Circ, output)
    suffix = "v" if output == "verilog" else "json"
    assert check_files_equal(__file__, f"build/{Circ.name}.{suffix}",
                             f"gold/{Circ.name}.{suffix}")


@pytest.mark.parametrize("width", [None, 3])
@pytest.mark.parametrize("output", ["verilog", "coreir"])
@pytest.mark.parametrize("op", [operator.imatmul, operator.le])
def test_assign_to_var(width, output, op):
    T = m.util.BitOrBits(width)

    class Circ(m.Circuit):
        name = f"test_assign_operator2_{width}_{output}"
        io = m.IO(a=m.In(T), b=m.In(T), c=m.Out(T))
        and2 = DeclareAnd(width)()
        c, I0, I1 = and2.I0, and2.I1, io.c
        op(I0, io.a)
        op(I1, io.b)
        op(c, and2.O)

    m.compile(f"build/{Circ.name}", Circ, output)
    suffix = "v" if output == "verilog" else "json"
    assert check_files_equal(__file__, f"build/{Circ.name}.{suffix}",
                             f"gold/{Circ.name}.{suffix}")


@pytest.mark.parametrize("width", [None, 3])
def test_assign_error_0(width):
    T = m.util.BitOrBits(width)

    class Circ(m.Circuit):
        name = f"test_assign_operator_{width}"
        io = m.IO(a=m.In(T), b=m.In(T), c=m.Out(T))
        and2 = DeclareAnd(width)()
        with pytest.raises(
                TypeError,
                match=rf"Cannot use @= to assign to output: O \(trying to assign a\)"):  # noqa
            and2.O @= io.a


@pytest.mark.parametrize("width", [None, 3])
def test_assign_error_1(width):
    T = m.util.BitOrBits(width)

    class Circ(m.Circuit):
        name = f"test_assign_operator_{width}"
        io = m.IO(a=m.In(T), b=m.In(T), c=m.Out(T))
        and2 = DeclareAnd(width)()
        with pytest.raises(
                TypeError,
                match=rf"Cannot use @= to assign to output: a \(trying to assign O\)"):  # noqa
            io.a @= and2.O


@pytest.mark.parametrize('T', [m.Bit, m.Array[5, m.Bit], m.Bits[5],
                               m.Tuple[m.Bit, m.Bits[5]]])
def test_eq_neq_input(T):
    class Foo(m.Circuit):
        io = m.IO(O=m.Out(T))
        with pytest.raises(TypeError, match="Cannot use == on an input"):
            io.O == 1
        with pytest.raises(TypeError, match="Cannot use != on an input"):
            io.O != 1



import pytest

import magma as m
import magma.testing
from magma.inline_verilog2 import InlineVerilogError


def _compile_and_check(top: m.DefineCircuitKind, **opts):
    basename = f"test_{top.name}"
    if opts.get('flatten_all_tuples', False):
        basename += "_flatten_all_tuples"
    m.compile(f"build/{basename}", top, output="mlir", **opts)
    assert m.testing.check_files_equal(
        __file__,
        f"build/{basename}.mlir",
        f"gold/{basename}.mlir",
    )


def test_simple():

    class _Top(m.Circuit):
        name = "inline_verilog2_simple"
        io = m.IO(I=m.In(m.Bit), O=m.Out(m.Bit), arr=m.In(m.Bits[2]))
        io += m.ClockIO()
        io.O <= m.register(io.I)
        m.inline_verilog2(
            "assert property (@(posedge CLK) {I} |-> ##1 {O});\n",
            O=io.O,
            I=io.I,
        )
        m.inline_verilog2(
            "assert property (@(posedge CLK) {io.arr[0]} |-> ##1 {io.arr[1]});\n",
        )

    _compile_and_check(_Top)


@pytest.mark.parametrize('flatten_all_tuples', [True, False])
def test_tuple(flatten_all_tuples):

    class RV(m.Product):
        data = m.In(m.Bits[5])
        valid = m.In(m.Bit)
        ready = m.Out(m.Bit)

    RVDATAIN = m.Array[2, RV]

    class InnerInnerDelayUnit(m.Circuit):
        name = "InnerInnerDelayUnit"
        io = m.IO(INPUT=RVDATAIN, OUTPUT=m.Flip(RVDATAIN))

    class InnerDelayUnit(m.Circuit):
        io = m.IO(INPUT=RVDATAIN, OUTPUT=m.Flip(RVDATAIN)) + m.ClockIO()

        delay = InnerInnerDelayUnit(name="inner_inner_delay")
        delay.INPUT[0] <= io.INPUT[1]
        delay.INPUT[1] <= io.INPUT[0]
        io.OUTPUT[0] <= delay.OUTPUT[1]
        io.OUTPUT[1] <= delay.OUTPUT[0]

    class DelayUnit(m.Circuit):
        io = m.IO(INPUT=RVDATAIN, OUTPUT=m.Flip(RVDATAIN)) + m.ClockIO()

        delay = InnerDelayUnit(name="inner_delay")
        delay.INPUT[0] <= io.INPUT[1]
        delay.INPUT[1] <= io.INPUT[0]
        io.OUTPUT[0] <= delay.OUTPUT[1]
        io.OUTPUT[1] <= delay.OUTPUT[0]

    class _Top(m.Circuit):
        io = m.IO(I=RVDATAIN, O=m.Flip(RVDATAIN)) + m.ClockIO()
        name = "inline_verilog2_tuple"

        delay = DelayUnit()
        delay.INPUT[0] <= io.I[1]
        delay.INPUT[1] <= io.I[0]
        io.O[1] <= delay.OUTPUT[0]
        io.O[0] <= delay.OUTPUT[1]

        assertion = (
            "assert property (@(posedge CLK) {valid_out} |-> ##3 {ready_out});"
        )

        m.inline_verilog2(
            assertion,
            valid_out=io.I[0].valid,
            ready_out=io.O[1].ready,
        )

        # Test inst ref.
        m.inline_verilog2(
            assertion,
            valid_out=delay.OUTPUT[1].valid,
            ready_out=delay.INPUT[0].ready,
        )

        # Test recursive ref.
        m.inline_verilog2(
            assertion,
            valid_out=delay.inner_delay.OUTPUT[0].valid,
            ready_out=delay.inner_delay.INPUT[1].ready,
        )

        # Test double recursive ref.
        m.inline_verilog2(
            assertion,
            valid_out=delay.inner_delay.inner_inner_delay.OUTPUT[0].valid,
            ready_out=delay.inner_delay.inner_inner_delay.INPUT[1].ready,
        )

    _compile_and_check(_Top, flatten_all_tuples=flatten_all_tuples)


def test_undriven_port_error():

    def gen():

        class _(m.Circuit):
            io = m.IO(I=m.In(m.Bit), O=m.Out(m.Bit), arr=m.In(m.Bits[2])) + m.ClockIO()
            # Expect an error because io.O is undriven.
            m.inline_verilog2(
                "assert property (@(posedge CLK) {I} |-> ##1 {O});",
                O=io.O, I=io.I,
            )

    with pytest.raises(InlineVerilogError) as e:
        gen()

    assert str(e.value) == "Found reference to undriven input port: LazyCircuit.O"


def test_uniquification():
    # Check that 2 modules that differ only in their inlined modules (e.g. one
    # displays in "%d" format, the other "%x") are uniquified.

    class Foo(m.Circuit):
        io = m.IO(I=m.In(m.Bit))
        m.inline_verilog2("always @(*) $display(\"%d\\n\", {io.I});")

    Bar = Foo

    class Foo(m.Circuit):
        io = m.IO(I=m.In(m.Bit))
        m.inline_verilog2("always @(*) $display(\"%x\\n\", {io.I});")

    class _Top(m.Circuit):
        name = "inline_verilog2_uniquification"
        io = m.IO(I=m.In(m.Bit))
        Bar()(io.I)
        Foo()(io.I)

    _compile_and_check(_Top)


def test_share_default_clocks():

    class _Top(m.Circuit):
        name = "inline_verilog2_share_default_clocks"
        io = m.IO(x=m.In(m.Bit), y=m.In(m.Bit)) + m.ClockIO(has_reset=True)

        clk = m.Clock()
        rst = m.Reset()
        m.inline_verilog2(
            "assert property (@(posedge {clk}) "
            "disable iff (! {rst}) {io.x} |-> ##1 {io.y});"
        )
        m.inline_verilog2(
            "assert property (@(posedge {clk}) "
            "disable iff (! {rst}) {io.x} |-> ##1 {io.y});"
        )

    _compile_and_check(_Top)


def test_passthrough_wire():

    class _Top(m.Circuit):
        name = "test_inline_verilog2_passthrough_wire"
        T = m.AnonProduct[dict(x=m.Bit, y=m.Bits[4])]
        io = m.IO(I=m.In(T), O=m.Out(T))
        io.O @= io.I

        m.inline_verilog2("assert {io.I.y[0]} == {io.I.y[1]}")
        m.inline_verilog2("assert {io.I.y[1:3]} == {io.I.y[2:4]}")

    _compile_and_check(_Top)


def test_clock_output():

    class _Top(m.Circuit):
        name = "test_inline_verilog2_clock_output"
        io = m.IO(x=m.In(m.Clock), y=m.In(m.Clock))
        m.inline_verilog2("Foo bar (.x({io.x}), .y({io.y}))")

    _compile_and_check(_Top)


def test_wire_insertion_bad_verilog():
    # See #1133 (https://github.com/phanrahan/magma/issues/1133).

    class _Top(m.Circuit):
        name = "test_inline_verilog2_wire_insertion_bad_verilog"
        io = m.IO(I=m.In(m.Bits[32]), O=m.Out(m.Bit))
        m.inline_verilog2("`ifdef LOGGING_ON")
        m.inline_verilog2("$display(\"%x\", {io.I[0]});")
        m.inline_verilog2("`endif LOGGING_ON")
        io.O @= io.I[0]

    _compile_and_check(_Top)



import magma as m


_SEQ2INT_CASES = (
    ([1, 0, 1, 1], 13),
    ([True, False, 1, 0], 5),
)
_INT2SEQ_CASES = (
    (14, None, [0, 1, 1, 1]),
    (12, 8, [0, 0, 1, 1, 0, 0, 0, 0]),
)
_CLOG2_CASES = (
    (1, 0),
    (7, 3),
    (14, 4),
    (16, 4),
)
_CLOG2SAFE_CASES = (
    (1, 1),
    (7, 3),
    (14, 4),
    (16, 4),
)


def test_seq2int():
    for s, i in _SEQ2INT_CASES:
        assert m.bitutils.seq2int(s) == i


def test_int2seq():
    for i, n, s in _INT2SEQ_CASES:
        assert m.bitutils.int2seq(i, n=n) == s


def test_clog2():
    for i, o in _CLOG2_CASES:
        assert m.bitutils.clog2(i) == o


def test_clog2safe():
    for i, o in _CLOG2SAFE_CASES:
        assert m.bitutils.clog2safe(i) == o



import contextlib
import itertools

from magma.backend.mlir.builtin import builtin
from magma.backend.mlir.hw import hw
from magma.backend.mlir.magma_common import (
    visit_value_or_value_wrapper_by_direction as
    visit_magma_value_or_value_wrapper_by_direction,
    tuple_key_to_str
)
from magma.backend.mlir.mlir import get_block_stack, MlirValue, push_block
from magma.backend.mlir.sv import sv
from magma.common import sort_by_value
from magma.primitives.when import iswhen
from magma.ref import TupleRef, DerivedRef
from magma.value_utils import make_selector


class _IndexBuilder:
    """Stages getitem/getattr calls by appending them to a tuple."""

    def __init__(self):
        self.index = tuple()

    def __getitem__(self, idx):
        self.index += (idx,)
        return self

    def __getattr__(self, idx):
        self.index += (idx,)
        return self


class WhenCompiler:
    def __init__(self, module_visitor, module):
        self._module_visitor = module_visitor
        self._module = module
        self._operands = self._module.operands
        self._flatten_all_tuples = (
            self._module_visitor.ctx.opts.flatten_all_tuples
        )

        # Track outer_block so array_ref/constants are placed outside the always
        # comb to reduce code repetition
        self._outer_block = get_block_stack().peek()

        inst = module.module
        defn = type(inst)
        assert iswhen(defn)
        self._builder = builder = defn._builder_

        # Update index map for flattened tuples
        self._input_to_index = self._flatten_index_map(builder.input_to_index)
        self._output_to_index = self._flatten_index_map(builder.output_to_index)

        self._output_wires = self._make_output_wires()

    def _get_input_index(self, value):
        return self._input_to_index[value]

    def _get_operand(self, value):
        return self._operands[self._get_input_index(value)]

    def _flatten_index_map(self, builder_map):
        """The when builder tracks a mapping from value to their index which is
        used to find a value's position in the input port list.  This function
        flattens tuple values and updates the index map accordingly since
        flattened tuples will offset the originally indices.
        """
        value_to_index = {}
        counter = itertools.count()

        def _index_map_visit(value):
            nonlocal counter, value_to_index
            if (
                self._flatten_all_tuples
                and isinstance(value.name, TupleRef)
                and value in value_to_index
            ):
                # tuples are flattened by the
                # `visit_magma_value_or_value_wrapper_by_direction`, so we avoid
                # adding them twice which invalidates the count logic
                return
            for ref in value.name.root_iter(
                stop_if=lambda ref: not isinstance(ref, DerivedRef)
            ):
                if ref.parent_value in value_to_index:
                    # Don't add, only need its parent
                    return

            # NOTE(leonardt): value may already be in value_to_index, in which
            # case we update it to a new index.  This is okay and it just means
            # that `value` occurs as more than one input (for example, it could
            # be used as a conditional driver for multiple values).  We could
            # optimize the logic to always share the same argument, but for now
            # we just use the "last" index.
            value_to_index[value] = next(counter)

        for value in sort_by_value(builder_map):
            visit_magma_value_or_value_wrapper_by_direction(
                value,
                _index_map_visit,
                _index_map_visit,
                flatten_all_tuples=self._flatten_all_tuples,
                inout_visitor=_index_map_visit
            )
        return value_to_index

    def _make_output_wires(self):
        """Create the mlir values corresponding to each output."""
        wires = [
            self._module_visitor.ctx.new_value(hw.InOutType(result.type))
            for result in self._module.results
        ]

        for result, wire in zip(self._module.results, wires):
            sv.RegOp(results=[wire])
            sv.ReadInOutOp(operands=[wire], results=[result])

        return wires

    def _flatten_value(self, value):
        fields = []
        visit_magma_value_or_value_wrapper_by_direction(
            value, fields.append, fields.append,
            flatten_all_tuples=self._flatten_all_tuples,
            inout_visitor=fields.append,
        )
        return fields

    def _get_parent(self, val, collection, to_index):
        """Search ancestor tree until we find either not an Array or Tuple or
        we find an array or tuple that is in the index map.
        """
        for ref in val.name.root_iter(
            stop_if=lambda ref: not isinstance(ref, DerivedRef)
        ):
            try:
                idx = to_index[ref.parent_value]
            except KeyError:
                pass  # try next parent
            else:
                return collection[idx], ref.parent_value
        return None, None  # didn't find parent

    def _check_derived_wire(self, val, collection, to_index):
        """If val is a child of an array or tuple in the index map, get the
        parent wire (so we add to a collection of drivers for a bulk assign).
        """
        wire, parent = self._get_parent(val, collection, to_index)
        if wire is None:
            return None, None

        builder = _IndexBuilder()
        make_selector(val, stop_at=parent).select(builder)
        return wire, builder.index

    def _make_operand(self, wire, index):
        """If the operand is an element of wire, emit the required array
        reference ops (extract, get, or slice) to retrieve the desired index."""
        if not index:
            return wire
        i = index[0]
        if isinstance(i, slice):
            # convert to tuple for hashing
            i = (i.start, i.stop, i.step)
        with push_block(self._outer_block):
            if isinstance(wire.type, (hw.ArrayType, builtin.IntegerType)):
                operand = self._module_visitor.make_array_ref(wire, i)
            else:
                assert isinstance(wire.type, hw.StructType)
                operand = self._module_visitor.make_struct_ref(wire, i)
        return self._make_operand(operand, index[1:])

    def _build_wire_map(self, connections):
        """Collect a map of output wires to their drivers.
        If it's an array that's been elaborated, we collect the drivers in a
        dictionary using their index as a key to sort.
        """
        wire_map = {}
        for drivee, driver in connections:
            elts = zip(*map(self._flatten_value, (drivee, driver)))
            for drivee_elt, driver_elt in elts:

                operand_wire, operand_index = self._check_derived_wire(
                    driver_elt, self._operands, self._input_to_index)
                if operand_wire:
                    operand = self._make_operand(operand_wire,
                                                 operand_index)
                else:
                    operand = self._get_operand(driver_elt)

                drivee_wire, drivee_index = self._check_derived_wire(
                    drivee_elt, self._output_wires, self._output_to_index)
                if drivee_wire:
                    wire_map.setdefault(drivee_wire, {})
                    drivee_index = tuple(
                        i if not isinstance(i, slice) else i.start
                        for i in drivee_index
                    )
                    wire_map[drivee_wire][drivee_index] = operand
                else:
                    drivee_idx = self._output_to_index[drivee_elt]
                    drivee_wire = self._output_wires[drivee_idx]
                    wire_map[drivee_wire] = operand
        return wire_map

    def _make_recursive_collection(self, T):
        """Create a nested data structure matching T, used to populate the
        elements of an array or struct create op.
        """
        if isinstance(T, builtin.IntegerType):
            return [None for _ in range(T.n)]
        if isinstance(T, hw.StructType):
            return {k: self._make_recursive_collection(v) for k, v in T.fields}
        assert isinstance(T, hw.ArrayType), T
        return [self._make_recursive_collection(T.T) for _ in range(T.dims[0])]

    def _populate_recursive_collection(self, T, value):
        """Unpack the contents of value into the corresponding nested structure
        create in `_make_recursive_collection`.
        """
        val = self._make_recursive_collection(T)
        for idx, elem in value.items():
            curr_val = val
            for i in range(len(idx)):
                curr_idx = idx[i]
                if isinstance(curr_val, dict) and isinstance(curr_idx, int):
                    curr_idx = f"_{curr_idx}"
                if i < len(idx) - 1:
                    curr_val = curr_val[curr_idx]  # descend until last index
                else:
                    curr_val[curr_idx] = elem  # use last index for setitem
        return val

    def _create_struct_from_collection(self, T, value, result):
        value = [
            self._create_from_recursive_collection(v, value[k])
            for k, v in T.fields
        ]
        hw.StructCreateOp(
            operands=value,
            results=[result]
        )
        return result

    def _create_from_recursive_collection(self, T, value):
        """Sort drivers by index, use concat or create depending on type."""
        if isinstance(value, MlirValue):
            return value  # found whole value, no need to combine
        if all(x is None for x in value):
            return None  # found empty value, covered by previous slice
        result = self._module_visitor.ctx.new_value(T)
        if not isinstance(T, builtin.IntegerType):
            # recursive combine children
            if isinstance(T, hw.StructType):
                return self._create_struct_from_collection(T, value, result)
            assert isinstance(T, hw.ArrayType)
            assert len(T.dims) == 1, "Expected 1d array"
            value = [
                self._create_from_recursive_collection(T.T, value[i])
                for i in range(T.dims[0])
            ]
        # Filter None elements (indices covered by a previous slice)
        value = [x for x in reversed(value) if x is not None]
        self._module_visitor.make_concat(value, result)
        return result

    def _make_assignments(self, connections):
        """
        * _build_wire_map: contructs mapping from output wire to driver

        * _populate_recursive_collection,
          _create_from_recursive_collection: handle collection of elaborated
                                             drivers for a bulk assign
        """
        for wire, value in self._build_wire_map(connections).items():
            if isinstance(value, dict):
                value = self._populate_recursive_collection(wire.type.T, value)
                value = self._create_from_recursive_collection(
                    wire.type.T, value
                )
            sv.BPAssignOp(operands=[wire, value])

    def _process_connections(self, block):
        connections = (
            (conditional_wire.drivee, conditional_wire.driver)
            for conditional_wire in block.conditional_wires()
        )
        self._make_assignments(connections)
        for child in block.children():
            self._process_when_block(child)

    def _process_when_block(self, block):
        """
        If no condition, we are in an otherwise case and simply emit the
        block body (which is inside a previous IfOp).

        Otherwise, we emit an IfOp with the true body corresponding to this
        block, then process the sibilings in the else block.
        """
        if block.condition is None:
            return self._process_connections(block)

        cond = self._get_operand(block.condition)
        if_op = sv.IfOp(operands=[cond])

        with push_block(if_op.then_block):
            self._process_connections(block)

        curr_sibling = if_op
        with contextlib.ExitStack() as stack:
            sibling_blocks = list(block.elsewhen_blocks())
            if block.otherwise_block is not None:
                sibling_blocks.append(block.otherwise_block)
            for sibling_block in sibling_blocks:
                stack.enter_context(push_block(curr_sibling.else_block))
                curr_sibling = self._process_when_block(sibling_block)

        return if_op

    def compile(self):
        """Emit default drivers then process the when block chain."""
        with push_block(sv.AlwaysCombOp().body_block):
            self._make_assignments(self._builder.default_drivers.items())
            self._process_when_block(self._builder.block)
        return True



import pytest

import magma as m
import magma.testing
import fault as f


def test_basic():

    class _Top(m.Circuit):
        io = m.IO(I=m.In(m.Bit), O=m.Out(m.Bit)) + m.ClockIO()

        with m.compile_guard("COND", defn_name="COND_compile_guard"):
            out = m.Register(m.Bit)()(io.I)

        io.O @= io.I

    m.compile("build/test_compile_guard_basic", _Top)
    assert m.testing.check_files_equal(
        __file__, f"build/test_compile_guard_basic.json",
        f"gold/test_compile_guard_basic.json")


def test_assert():

    class _Top(m.Circuit):
        io = m.IO(I=m.In(m.Valid[m.Bits[4]]), O=m.Out(m.Bits[4])) + m.ClockIO()
        io.O @= m.Register(m.Bits[4])()(io.I.data)

        with m.compile_guard("ASSERT_ON", "ASSERT_ON_compile_guard"):
            count = m.Register(m.UInt[2], has_enable=True)()
            count.I @= count.O + 1
            count.CE @= io.I.valid
            f.assert_immediate((count.O != 3) | (io.O.value() == 3))

    m.compile("build/test_compile_guard_assert", _Top, inline=True)
    assert m.testing.check_files_equal(
        __file__, f"build/test_compile_guard_assert.json",
        f"gold/test_compile_guard_assert.json")


def test_array():

    class _Top(m.Circuit):
        io = m.IO(I=m.In(m.Array[2, m.Bit]), O=m.Out(m.Bit)) + m.ClockIO()

        with m.compile_guard("COND", defn_name="COND_compile_guard"):
            out = m.Register(m.Bit)()(io.I[0])

        io.O @= io.I[1]

    m.compile("build/test_compile_guard_array", _Top)
    assert m.testing.check_files_equal(
        __file__, f"build/test_compile_guard_array.json",
        f"gold/test_compile_guard_array.json")


def test_multiple_array():

    class _Top(m.Circuit):
        io = m.IO(I=m.In(m.Array[2, m.Bit]), O=m.Out(m.Bit)) + m.ClockIO()

        with m.compile_guard("COND", defn_name="COND_compile_guard"):
            m.Register(m.Bit)()(io.I[1])
            m.Register(m.Bit)()(io.I[0])

        io.O @= io.I[1]

    m.compile("build/test_compile_guard_multiple_array", _Top)
    assert m.testing.check_files_equal(
        __file__, f"build/test_compile_guard_multiple_array.json",
        f"gold/test_compile_guard_multiple_array.json")


def test_nested_type():

    class _Top(m.Circuit):
        T = m.Product.from_fields("anon", dict(x=m.Bit, y=m.Bit))
        T = m.Array[2, T]

        io = m.IO(I=m.In(T), O=m.Out(m.Bit)) + m.ClockIO()

        with m.compile_guard("COND", defn_name="COND_compile_guard"):
            m.Register(m.Bit)()(io.I[1].x)
            m.Register(m.Bit)()(io.I[0].y)

        io.O @= io.I[0].x

    m.compile("build/test_compile_guard_nested_type", _Top)
    assert m.testing.check_files_equal(
        __file__, f"build/test_compile_guard_nested_type.json",
        f"gold/test_compile_guard_nested_type.json")


def test_anon_driver_driven():

    def make_top():

        class _Top(m.Circuit):
            io = m.IO(I=m.In(m.Bit))
            x = m.Bit()
            x @= io.I
            with m.compile_guard("COND", defn_name="A"):
                m.register(x)

        return _Top

    Top = make_top()
    basename = "test_compile_guard_anon_driver_driven"
    m.compile(f"build/{basename}", Top)
    assert m.testing.check_files_equal(
        __file__, f"build/{basename}.v", f"gold/{basename}.v"
    )


def test_anon_driver_undriven():

    def make_top():

        class _Top(m.Circuit):
            io = m.IO(I=m.In(m.Bit))
            x = m.Bit()
            with m.compile_guard("COND", defn_name="A"):
                m.register(x)

        return _Top

    Top = make_top()
    basename = "test_compile_guard_anon_driver_undriven"
    with pytest.raises(m.MagmaCompileException):
        m.compile(f"build/{basename}", Top)


def test_anon_driver_nested_type():

    def make_top():

        class _Top(m.Circuit):
            io = m.IO(I=m.In(m.Bit))
            T = m.Array[10, m.AnonProduct[{"x": m.Bits[10]}]]
            x = T()
            for x_i in x:
                for x_i_j in x_i.x:
                    x_i_j @= io.I
            with m.compile_guard("COND", defn_name="A"):
                m.register(x)

        return _Top

    Top = make_top()
    basename = "test_compile_guard_anon_driver_nested_type"
    m.compile(f"build/{basename}", Top)
    assert m.testing.check_files_equal(
        __file__, f"build/{basename}.v", f"gold/{basename}.v"
    )


@pytest.mark.skip(reason="nested compile guard context not yet implemented")
def test_nested_context():

    class _Top(m.Circuit):
        io = m.IO(I0=m.In(m.Bit), I1=m.In(m.Bit), O=m.Out(m.Bit)) + m.ClockIO()

        with m.compile_guard("OUTER", defn_name="OUTER_compile_guard"):
            m.Register(m.Bit)()(io.I0)

            with m.compile_guard("INNER", defn_name="INNER_compile_guard"):
                m.Register(m.Bit)()(io.I1)

        io.O @= io.I0

    m.compile("build/test_compile_guard_nested_context", _Top)
    assert m.testing.check_files_equal(
        __file__, f"build/test_compile_guard_nested_context.json",
        f"gold/test_compile_guard_nested_context.json")


def test_basic_oldstyle():

    class _Top(m.Circuit):
        IO = ["I", m.In(m.Bit), "O", m.Out(m.Bit)] + m.ClockInterface()

        @classmethod
        def definition(io):
            with m.compile_guard("COND", defn_name="COND_compile_guard"):
                out = m.Register(m.Bit)()(io.I)

            io.O @= io.I

    m.compile("build/test_compile_guard_basic", _Top)
    assert m.testing.check_files_equal(
        __file__, f"build/test_compile_guard_basic.json",
        f"gold/test_compile_guard_basic.json")


def test_basic_undefined():

    class _Top(m.Circuit):
        io = m.IO(I=m.In(m.Bit), O=m.Out(m.Bit)) + m.ClockIO()

        with m.compile_guard("COND", defn_name="COND_compile_guard",
                             type='undefined'):
            out = m.Register(m.Bit)()(io.I)

        io.O @= io.I

    m.compile("build/test_compile_guard_basic_undefined", _Top)
    assert m.testing.check_files_equal(
        __file__, f"build/test_compile_guard_basic_undefined.json",
        f"gold/test_compile_guard_basic_undefined.json")


def test_vcc():

    class _Top(m.Circuit):
        io = m.IO(I=m.In(m.Bit), O=m.Out(m.Bit)) + m.ClockIO()

        with m.compile_guard("COND", defn_name="COND_compile_guard"):
            out = m.Register(m.Bit)()(io.I ^ 1)

        io.O @= io.I

    m.compile("build/test_compile_guard_basic_vcc", _Top)
    assert m.testing.check_files_equal(
        __file__, f"build/test_compile_guard_basic_vcc.json",
        f"gold/test_compile_guard_basic_vcc.json")


def test_drive_outputs():

    class _Top(m.Circuit):
        io = m.IO(I=m.In(m.Bit), O=m.Out(m.Bit)) + m.ClockIO()

        with pytest.raises(TypeError):
            with m.compile_guard("COND"):
                io.O @= m.Register(m.Bit)()(io.I ^ 1)


def test_anon_drivee():

    def make_top():
        class _Top(m.Circuit):
            io = m.IO(I=m.In(m.Bit))
            O = m.Bit()

            with m.compile_guard("COND"):
                O @= m.register(io.I)

        return _Top

    with pytest.raises(NotImplementedError):
        _ = make_top()


def test_compile_guard_select_basic():
    class _Top(m.Circuit):
        io = m.IO(I=m.In(m.Bit), O=m.Out(m.Bit)) + m.ClockIO()

        x = m.Register(m.Bit)()(io.I ^ 1)
        y = m.Register(m.Bit)()(io.I)

        io.O @= m.compile_guard_select(
            COND1=x, COND2=y, default=io.I
        )

    basename = "test_compile_guard_select_basic"
    m.compile(f"build/{basename}", _Top, inline=True)
    assert m.testing.check_files_equal(
        __file__, f"build/{basename}.v", f"gold/{basename}.v")


def test_compile_guard_select_complex_type():
    T = m.Product.from_fields("anonymous", dict(x=m.Bit, y=m.Bit))

    def make_top():

        class _Top(m.Circuit):
            io = m.IO(I0=m.In(T), I1=m.In(T), O=m.Out(T))
            io.O @= m.compile_guard_select(
                COND1=io.I0, COND2=io.I1, default=io.I0)

    with pytest.raises(TypeError):
        make_top()


def test_contained_inline_verilog():

    class Top(m.Circuit):
        io = m.IO(I=m.In(m.Bit), O=m.Out(m.Bit))
        io.O @= io.I
        with m.compile_guard("DEBUG", "DebugModule"):
            reg = m.Register(m.Bit)(name="reg")
            reg.I @= reg.O | io.I
            m.inline_verilog("assert {io.I};")
            m.inline_verilog("assert ~{reg.O};")

    basename = "test_compile_guard_contained_inline_verilog"
    m.compile(f"build/{basename}", Top)
    assert m.testing.check_files_equal(
        __file__, f"build/{basename}.v", f"gold/{basename}.v")


def test_compile_guard_anon_driven_internal():

    class _Top(m.Circuit):
        io = m.IO(I=m.In(m.Bit), O=m.Out(m.Bit)) + m.ClockIO()

        with m.compile_guard("COND", defn_name="COND_compile_guard"):
            x = m.Bit(name="x")
            x.undriven()  # x is driven internally
            out = m.Register(m.Bit)()(x)

        io.O @= io.I

    basename = "test_compile_guard_anon_driven_internal"
    m.compile(f"build/{basename}", _Top, output="mlir")
    assert m.testing.check_files_equal(
        __file__, f"build/{basename}.mlir", f"gold/{basename}.mlir"
    )


def test_compile_guard_anon_driving_external():

    class _Top(m.Circuit):
        io = m.IO(I=m.In(m.Bit), O=m.Out(m.Bit)) + m.ClockIO()
        x = m.Bit(name="x")
        io.O @= x  # x drives an external value

        with pytest.raises(TypeError):
            with m.compile_guard("COND", defn_name="COND_compile_guard"):
                x @= m.Register(m.Bit)()(io.I)


def test_compile_guard_inline_verilog_reset():

    class _Top(m.Circuit):
        io = m.IO(I=m.In(m.Bit), O=m.Out(m.Bit))
        io += m.ClockIO(has_async_resetn=True)
        io.O @= io.I

        with m.compile_guard("COND", defn_name="COND_compile_guard"):
            m.inline_verilog(
                """
assert property (@(posedge {clk}) disable iff {rst} {x} |-> ##1 {x};
                """,
                x=io.I,
                rst=m.AsyncResetN(),
                clk=m.Clock()
            )

    basename = "test_compile_guard_inline_verilog_reset"
    m.compile(f"build/{basename}", _Top, output="mlir")
    assert m.testing.check_files_equal(
        __file__, f"build/{basename}.mlir", f"gold/{basename}.mlir")



import pathlib
from typing import Any, Dict, Iterable
import weakref

from magma.backend.mlir.builtin import builtin
from magma.backend.mlir.compile_to_mlir_opts import CompileToMlirOpts
from magma.backend.mlir.errors import MlirCompilerInternalError
from magma.backend.mlir.hardware_module import HardwareModule
from magma.backend.mlir.hw import hw
from magma.backend.mlir.mlir import MlirOp, MlirSymbol, push_block
from magma.backend.mlir.mlir_passes import CollectMlirOpsPass
from magma.backend.mlir.scoped_name_generator import ScopedNameGenerator
from magma.backend.mlir.sv import sv
from magma.bind2 import is_bound_module
from magma.circuit import CircuitKind, DefineCircuitKind
from magma.common import find_by_value, only
from magma.passes import dependencies
from magma.t import Type


def _set_module_attrs(mlir_module: builtin.ModuleOp, opts: CompileToMlirOpts):
    lowering_options = [
        f"locationInfoStyle={opts.location_info_style}",
    ]
    if opts.explicit_bitcast:
        lowering_options.append("explicitBitcast")
    if opts.disallow_expression_inlining_in_ports:
        lowering_options.append("disallowExpressionInliningInPorts")
    if opts.disallow_local_variables:
        lowering_options.append("disallowLocalVariables")
    if opts.omit_version_comment:
        lowering_options.append("omitVersionComment")
    if opts.disallow_packed_struct_assignments:
        lowering_options.append("disallowPackedStructAssignments")
    if lowering_options:
        mlir_module.attr_dict["circt.loweringOptions"] = builtin.StringAttr(
            f"{','.join(lowering_options)}"
        )


def _prepare_for_split_verilog(
        hardware_modules: Iterable[HardwareModule],
        bind_ops: Iterable[sv.BindOp],
        symbol_map: Dict[Any, MlirSymbol],
        basename: str,
        suffix: str,
):
    filename = pathlib.Path(f"{basename}.{suffix}")
    bound_module_to_filename = {}
    for hardware_module in hardware_modules:
        if hardware_module.hw_module is None:
            continue
        magma_defn_or_decl = hardware_module.magma_defn_or_decl
        if is_bound_module(magma_defn_or_decl):
            name = magma_defn_or_decl.name
            bound_module_to_filename[magma_defn_or_decl] = output_filename = (
                filename.parent / f"{name}.{suffix}"
            )
        else:
            output_filename = filename
        hardware_module.hw_module.attr_dict["output_file"] = (
            hw.OutputFileAttr(str(output_filename))
        )
    for bind_op in bind_ops:
        magma_inst = only(find_by_value(symbol_map, bind_op.instance.name))
        output_filename = str(bound_module_to_filename[type(magma_inst)])
        curr_op = bind_op
        while True:
            curr_op.attr_dict["output_file"] = (
                hw.OutputFileAttr(output_filename)
            )
            curr_op = curr_op.parent_op()
            if curr_op is None or not isinstance(curr_op, sv.IfDefOp):
                break


class TranslationUnit:
    def __init__(self, magma_top: DefineCircuitKind, opts: CompileToMlirOpts):
        self._magma_top = magma_top
        self._opts = opts
        self._mlir_module = builtin.ModuleOp()
        _set_module_attrs(self._mlir_module, opts)
        self._hardware_modules = {}
        self._symbol_map = {}
        self._symbol_name_generator = ScopedNameGenerator(
            disallow_duplicates=self._opts.disallow_duplicate_symbols,
        )
        self._external_bind_files = []
        self._xmr_paths = {}

    @property
    def magma_top(self) -> DefineCircuitKind:
        return self._magma_top

    @property
    def mlir_module(self) -> builtin.ModuleOp:
        return self._mlir_module

    @property
    def xmr_paths(self) -> Dict:
        return self._xmr_paths

    def new_hardware_module(
            self, magma_defn_or_decl: CircuitKind) -> HardwareModule:
        return HardwareModule(magma_defn_or_decl, weakref.ref(self), self._opts)

    def get_hardware_module(
            self, magma_defn_or_decl: CircuitKind) -> HardwareModule:
        key = self._make_key(magma_defn_or_decl)
        return self._hardware_modules[key]

    def set_hardware_module(
            self, magma_defn_or_decl: CircuitKind,
            hardware_module: HardwareModule):
        key = self._make_key(magma_defn_or_decl)
        if key in self._hardware_modules:
            raise ValueError(f"Hardware module '{key}' already exists")
        self._hardware_modules[key] = hardware_module

    def has_hardware_module(
            self, magma_defn_or_decl: CircuitKind) -> bool:
        key = self._make_key(magma_defn_or_decl)
        return key in self._hardware_modules

    def get_mapped_symbol(self, obj: Type) -> MlirSymbol:
        return self._symbol_map[obj]

    def get_or_make_mapped_symbol(self, obj: Any, **kwargs) -> MlirSymbol:
        try:
            return self._symbol_map[obj]
        except KeyError:
            pass
        self._symbol_map[obj] = symbol = self.new_symbol(**kwargs)
        return symbol

    def set_mapped_symbol(self, obj: Any, symbol: MlirSymbol):
        if obj in self._symbol_map:
            raise ValueError(f"{obj} already mapped")
        self._symbol_map[obj] = symbol

    def new_symbol(self, **kwargs) -> MlirSymbol:
        name = self._symbol_name_generator(**kwargs)
        return MlirSymbol(name)

    def add_external_bind_file(self, filename: str):
        self._external_bind_files.append(filename)

    def compile(self):
        deps = dependencies(self._magma_top, include_self=True)
        with push_block(self._mlir_module):
            for dep in deps:
                if self.has_hardware_module(dep):
                    continue
                hardware_module = self.new_hardware_module(dep)
                hardware_module.compile()
                if hardware_module.hw_module:
                    self.set_hardware_module(dep, hardware_module)
        if self._opts.split_verilog:
            if self._opts.basename is None:
                raise ValueError(
                    "Must specify basename if split_verilog is set"
                )
            _prepare_for_split_verilog(
                self._hardware_modules.values(),
                filter(
                    lambda op: isinstance(op, sv.BindOp),
                    CollectMlirOpsPass(self._mlir_module).run(),
                ),
                self._symbol_map,
                self._opts.basename,
                ("sv" if self._opts.sv else "v"),
            )
        self._process_bound_modules()

    @staticmethod
    def _make_key(magma_defn_or_decl: CircuitKind) -> str:
        return magma_defn_or_decl.name

    def _process_bound_modules(self):
        filelist_filename = self._opts.basename + "_bind_files.list"
        has_native_bound_modules = False
        for hardware_module in self._hardware_modules.values():
            if hardware_module.hw_module is None:
                continue
            if not is_bound_module(hardware_module.magma_defn_or_decl):
                continue
            has_native_bound_modules = True
            hardware_module.hw_module.attr_dict["output_filelist"] = (
                hw.FileListAttr(filelist_filename)
            )
        if not self._external_bind_files:
            return
        if has_native_bound_modules:
            raise MlirCompilerInternalError(
                "Mix of native and external (likely CoreIR compiled) bind "
                "modules not supported."
            )
        with open(filelist_filename, "w") as f:
            f.write("\n".join(self._external_bind_files))



from typing import Optional

from magma.bit import Bit
from magma.bits import UInt
from magma.clock_io import gen_clock_io
from magma.generator import Generator2
from magma.interface import IO
from magma.mantle.util import ispow2
from magma.primitives.mux import mux
from magma.primitives.register import Register
from magma.t import Type, Out


class Counter(Generator2):
    """Counts to `n` (0 to n - 1) repeatedly.
    """

    def __init__(
        self,
        n: int,
        has_enable: bool = False,
        has_cout: bool = False,
        reset_type: Optional[Type] = None
    ):
        num_bits = max((n - 1).bit_length(), 1)

        self.io = IO(O=Out(UInt[num_bits]))
        if has_cout:
            self.io += IO(COUT=Out(Bit))
        self.io += gen_clock_io(reset_type, has_enable)

        reg = Register(
            UInt[num_bits], has_enable=has_enable, reset_type=reset_type
        )()

        if has_enable:
            reg.CE @= self.io.CE

        self.io.O @= reg.O

        if has_cout:
            COUT = reg.O == (n - 1)
            if has_enable:
                COUT &= self.io.CE
            self.io.COUT @= COUT

        I = reg.O + 1
        if not ispow2(n):
            I = mux([I, 0], reg.O == (n - 1))
        reg.I @= I



from magma.passes.passes import DefinitionPass, pass_lambda
from magma.passes.split_when_utils import split_when_cycles
from magma.primitives.when import WhenBuilder
from magma.when import WhenBlock


class WhenPass(DefinitionPass):
    def __call__(self, definition):
        with definition.open():
            for builder in list(definition._context_._builders):
                if isinstance(builder, WhenBuilder):
                    self.process_when_builder(builder, definition)


class InferLatches(WhenPass):
    def process_when_builder(self, builder, defn):
        builder.infer_latches()


class EmitWhenAsserts(WhenPass):
    def __init__(self, main, flatten_all_tuples):
        super().__init__(main)
        self._flatten_all_tuples = flatten_all_tuples

    def process_when_builder(self, builder, defn):
        builder.emit_when_assertions(self._flatten_all_tuples)


class SplitCycles(WhenPass):
    def process_when_builder(self, builder, defn):
        split_when_cycles(builder, defn)


class FinalizeWhens(WhenPass):
    def process_when_builder(self, builder, defn):
        assert not builder._finalized
        defn._context_._placer.place(builder.finalize())


infer_latch_pass = pass_lambda(InferLatches)
emit_when_assert_pass = pass_lambda(EmitWhenAsserts)
split_cycles_pass = pass_lambda(SplitCycles)
finalize_when_pass = pass_lambda(FinalizeWhens)


def run_when_passes(
        main,
        flatten_all_tuples: bool = False,
        emit_when_assertions: bool = False
):
    infer_latch_pass(main)
    if emit_when_assertions:
        emit_when_assert_pass(main, flatten_all_tuples)
    split_cycles_pass(main)
    finalize_when_pass(main)



import random

import pytest

import magma as m

WIDTH = 4
NTESTS = 8

@pytest.mark.parametrize('T', [m.Bits, m.UInt, m.SInt])
def test_zext(T):
    for _ in range(NTESTS):
        val = random.randrange(1 << WIDTH)
        ext = random.randrange(1, 5)
        a =  T[WIDTH](val)
        b = a.zext(ext)
        assert b.bits() == a.concat(T[ext](0)).bits()


@pytest.mark.parametrize('T', [m.SInt])
def test_sext(T):
    for _ in range(NTESTS):
        val = random.randrange(1 << WIDTH)
        ext = random.randrange(1, 5)
        a =  T[WIDTH](val)
        b = a.sext(ext)
        if (val >> (WIDTH - 1)):
            assert b.bits() == a.concat(T[ext](-1)).bits()
        else:
            assert b.bits() == a.concat(T[ext](0)).bits()


@pytest.mark.parametrize('T', [m.Bits, m.UInt, m.SInt])
def test_ext(T):
    for _ in range(NTESTS):
        val = random.randrange(1 << WIDTH)
        ext = random.randrange(1, 5)
        a =  T[WIDTH](val)
        b = a.ext(ext)
        if T is m.SInt:
            assert b.bits() == a.sext(ext).bits()
        else:
            assert b.bits() == a.zext(ext).bits()


@pytest.mark.parametrize('T', [m.Bits, m.UInt, m.SInt])
def test_types(T):
    a = T[WIDTH]()
    ext = random.randrange(1, 5)
    b = a.zext(ext)
    assert isinstance(b, T)
    if T is m.SInt:
        c = a.sext(ext)
        assert isinstance(c, T)


def test_ext_n():
    value = m.SInt[32]()

    # zext TypeError.
    with pytest.raises(TypeError) as pytest_e:
        m.zext(value, None)
    assert pytest_e.type is TypeError
    assert pytest_e.value.args == ("Expected non-negative integer, got 'None'",)

    # zext ValueError.
    with pytest.raises(TypeError) as pytest_e:
        m.zext(value, -1)
    assert pytest_e.type is TypeError
    assert pytest_e.value.args == ("Expected non-negative integer, got '-1'",)

    # zext identity.
    assert m.zext(value, 0) is value

    # sext TypeError.
    with pytest.raises(TypeError) as pytest_e:
        m.sext(value, None)
    assert pytest_e.type is TypeError
    assert pytest_e.value.args == ("Expected non-negative integer, got 'None'",)

    # sext ValueError.
    with pytest.raises(TypeError) as pytest_e:
        m.sext(value, -1)
    assert pytest_e.type is TypeError
    assert pytest_e.value.args == ("Expected non-negative integer, got '-1'",)

    # sext identity.
    assert m.sext(value, 0) is value



from hwtypes.bit_vector_abc import TypeFamily
import magma as m


def test_basic():
    assert m.get_family() is m._Family_
    family = m.get_family()
    assert isinstance(family, TypeFamily)
    assert family.Bit is m.Bit
    assert family.BitVector is m.Bits
    assert family.Unsigned is m.UInt
    assert family.Signed is m.SInt
    for t in (m.Bit, m.Bits[1], m.UInt[1], m.SInt[1]):
        assert t.get_family() is family
